# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for plotting outputs of Torax runs."""

from collections.abc import Sequence
import dataclasses
import enum
from os import path
from typing import Any, List

import matplotlib
from matplotlib import gridspec
from matplotlib import widgets
import matplotlib.pyplot as plt
import numpy as np
from torax._src.output_tools import output
import xarray as xr

# Internal import.

# Constants for figure setup, plot labels, and formatting.
# The axes are designed to be plotted in the order they appear in the list,
# first ascending in columns, then rows.


class PlotType(enum.Enum):
  """Enum for plot types.

  SPATIAL: Spatial plots, e.g., 1D profiles as a function of toroidal flux
  coordinate. Plots change with time, and are modified by the slider.
  TIME_SERIES: Time series plots. 0D profiles plotting as a function of time.
  These plots are not modified by the slider.
  """

  SPATIAL = 1
  TIME_SERIES = 2


@dataclasses.dataclass
class PlotProperties:
  """Dataclass for individual plot properties."""

  attrs: tuple[str, ...]
  labels: tuple[str, ...]
  ylabel: str
  legend_fontsize: int | None = None  # None reverts to default matplotlib value
  upper_percentile: float = 100.0
  lower_percentile: float = 0.0
  include_first_timepoint: bool = True
  ylim_min_zero: bool = True
  plot_type: PlotType = PlotType.SPATIAL
  suppress_zero_values: bool = False  # If True, all-zero-data is not plotted


@dataclasses.dataclass
class FigureProperties:
  """Dataclass for all figure related data."""

  rows: int
  cols: int
  axes: tuple[PlotProperties, ...]
  figure_size_factor: float = 5.0
  tick_fontsize: int = 10
  axes_fontsize: int = 10
  title_fontsize: int = 16
  default_legend_fontsize: int = 10
  colors: tuple[str, ...] = ('r', 'b', 'g', 'm', 'y', 'c')

  def __post_init__(self):
    if len(self.axes) > self.rows * self.cols:
      raise ValueError('len(axes) in plot_config is more than rows * columns.')

  @property
  def contains_spatial_plot_type(self) -> bool:
    """Checks if any plot is a spatial plottype."""
    return any(
        plot_properties.plot_type == PlotType.SPATIAL
        for plot_properties in self.axes
    )


# pylint: disable=invalid-name
@dataclasses.dataclass
class PlotData:
  r"""Dataclass for all plot related data.

  Attributes:
    T_i: Ion temperature profile [:math:`\mathrm{keV}`] on the cell grid.
    T_e: Electron temperature profile [:math:`\mathrm{keV}`] on the cell grid.
    n_e: Electron density profile [:math:`\mathrm{10^{20} m^{-3}}`] on the cell
      grid.
    n_i: Main ion density profile [:math:`\mathrm{10^{20} m^{-3}}`] on the cell
      grid. Corresponds to a bundled ion mixture if specified as such in the
      config.
    n_impurity: Impurity density profile [:math:`\mathrm{10^{20} m^{-3}}`] on
      the cell grid. Corresponds to a bundled ion mixture if specified as such
      in the config.
    Z_impurity: Average charge state of the impurity species [dimensionless] on
      the cell grid.
    psi: Poloidal flux [:math:`\mathrm{Wb}`] on the cell grid.
    v_loop: Time derivative of poloidal flux (loop voltage :math:`V_{loop}`) [V]
      on the cell grid.
    j_total: Total toroidal current density profile [:math:`\mathrm{MA/m^2}`] on
      the cell grid.
    j_ohmic: Ohmic current density profile [:math:`\mathrm{MA/m^2}`] on the cell
      grid.
    j_bootstrap: Bootstrap current density profile [:math:`\mathrm{MA/m^2}`] on
      the cell grid.
    j_ecrh: Electron cyclotron current density profile [:math:`\mathrm{MA/m^2}`]
      on the cell grid.
    j_generic_current: Generic external current source density profile
      [:math:`\mathrm{MA/m^2}`] on the cell grid.
    j_external: Total externally driven current source density profile
      [:math:`\mathrm{MA/m^2}`] on the cell grid.
    q: Safety factor (q-profile) [dimensionless] on the face grid.
    magnetic_shear: Magnetic shear profile [dimensionless] on the face grid.
    chi_turb_i: Ion heat conductivity [:math:`\mathrm{m^2/s}`] on the face grid.
    chi_turb_e: Electron heat conductivity [:math:`\mathrm{m^2/s}`] on the face
      grid.
    D_turb_e: Electron particle diffusivity [:math:`\mathrm{m^2/s}`] on the face
      grid.
    V_turb_e: Electron particle convection velocity [:math:`\mathrm{m/s}`] on
      the face grid.
    p_icrh_i: ICRH ion heating power density [:math:`\mathrm{MW/m^3}`].
    p_icrh_e: ICRH electron heating power density [:math:`\mathrm{MW/m^3}`].
    p_generic_heat_i: Generic ion heating power density
      [:math:`\mathrm{MW/m^³}`].
    p_generic_heat_e: Generic electron heating power density
      [:math:`\mathrm{MW/m^3}`].
    p_ecrh_e: Electron cyclotron heating power density
      [:math:`\mathrm{MW/m^3}`].
    p_alpha_i: Fusion alpha particle heating power density to ion
      [:math:`\mathrm{MW/m^3}`].
    p_alpha_e: Fusion alpha particle heating power density to electrons
      [:math:`\mathrm{MW/m^3}`].
    p_ohmic_e: Ohmic heating power density [:math:`\mathrm{MW/m^3}`].
    p_bremsstrahlung_e: Bremsstrahlung radiation power density
      [:math:`\mathrm{MW/m^3}`].
    p_cyclotron_radiation_e: Cyclotron radiation power density
      [:math:`\mathrm{MW/m^3}`].
    ei_exchange: Ion-electron heat exchange power density
      [:math:`\mathrm{MW/m^3}`]. Positive values denote ion heating, and
      negative values denote ion cooling (electron heating).
    p_impurity_radiation_e: Impurity radiation power density
      [:math:`\mathrm{MW/m^3}`].
    Q_fusion: Fusion power gain (dimensionless).
    s_gas_puff: Gas puff particle source density [:math:`\mathrm{10^{20} m^{-3}
      s^{-1}}`].
    s_generic_particle: Generic particle source density [:math:`\mathrm{10^{20}
      m^{-3} s^{-1}}`].
    s_pellet: Pellet particle source density [:math:`\mathrm{10^{20} m^{-3}
      s^{-1}}`].
    Ip_profile: Total plasma current [:math:`\mathrm{MA}`].
    I_bootstrap: Total bootstrap current [:math:`\mathrm{MA}`].
    I_aux_generic: Total generic current source [:math:`\mathrm{MA}`].
    I_ecrh: Total electron cyclotron current [:math:`\mathrm{MA}`].
    P_auxiliary: Total auxiliary heating power [:math:`\mathrm{MW}`].
    P_ohmic_e: Total Ohmic heating power [:math:`\mathrm{MW}`].
    P_alpha_total: Total fusion alpha heating power [:math:`\mathrm{MW}`].
    P_sink: Total electron heating sink [:math:`\mathrm{MW}`].
    P_bremsstrahlung_e: Total bremsstrahlung radiation power loss
      [:math:`\mathrm{MW}`].
    P_cyclotron_e: Total cyclotron radiation power loss [:math:`\mathrm{MW}`].
    P_radiation_e: Total impurity radiation power loss [:math:`\mathrm{MW}`].
    t: Simulation time [:math:`\mathrm{s}`].
    rho_norm: Normalized toroidal flux coordinate on cell grid + boundaries.
    rho_cell_norm: Normalized toroidal flux coordinate on the cell grid.
    rho_face_norm: Normalized toroidal flux coordinate on the face grid.
    T_e_volume_avg: Volume-averaged electron temperature [:math:`\mathrm{keV}`].
    T_i_volume_avg: Volume-averaged ion temperature [:math:`\mathrm{keV}`].
    n_e_volume_avg: Volume-averaged electron density [:math:`\mathrm{10^{20}
      m^{-3}}`].
    n_i_volume_avg: Volume-averaged ion density [:math:`\mathrm{10^{20}
      m^{-3}}`].
    W_thermal_total: Total thermal stored energy [:math:`\mathrm{MJ}`].
    q95: Safety factor at 95% of the normalized poloidal flux.
  """

  T_i: np.ndarray
  T_e: np.ndarray
  n_e: np.ndarray
  n_i: np.ndarray
  n_impurity: np.ndarray
  Z_impurity: np.ndarray
  psi: np.ndarray
  v_loop: np.ndarray
  j_total: np.ndarray
  j_ohmic: np.ndarray
  j_bootstrap: np.ndarray
  j_ecrh: np.ndarray
  j_generic_current: np.ndarray
  j_external: np.ndarray
  q: np.ndarray
  magnetic_shear: np.ndarray
  chi_turb_i: np.ndarray
  chi_turb_e: np.ndarray
  D_turb_e: np.ndarray
  V_turb_e: np.ndarray
  p_icrh_i: np.ndarray
  p_icrh_e: np.ndarray
  p_generic_heat_i: np.ndarray
  p_generic_heat_e: np.ndarray
  p_ecrh_e: np.ndarray
  p_alpha_i: np.ndarray
  p_alpha_e: np.ndarray
  p_ohmic_e: np.ndarray
  p_bremsstrahlung_e: np.ndarray
  p_cyclotron_radiation_e: np.ndarray
  ei_exchange: np.ndarray
  p_impurity_radiation_e: np.ndarray
  Q_fusion: np.ndarray  # pylint: disable=invalid-name
  s_gas_puff: np.ndarray
  s_generic_particle: np.ndarray
  s_pellet: np.ndarray
  Ip_profile: np.ndarray
  I_bootstrap: np.ndarray
  I_aux_generic: np.ndarray
  I_ecrh: np.ndarray
  P_auxiliary: np.ndarray
  P_ohmic_e: np.ndarray
  P_alpha_total: np.ndarray
  P_sink: np.ndarray
  P_bremsstrahlung_e: np.ndarray
  P_cyclotron_e: np.ndarray
  P_radiation_e: np.ndarray
  t: np.ndarray
  rho_norm: np.ndarray
  rho_cell_norm: np.ndarray
  rho_face_norm: np.ndarray
  T_e_volume_avg: np.ndarray
  T_i_volume_avg: np.ndarray
  n_e_volume_avg: np.ndarray
  n_i_volume_avg: np.ndarray
  W_thermal_total: np.ndarray  # pylint: disable=invalid-name
  q95: np.ndarray


def load_data(filename: str) -> PlotData:
  """Loads an xr.Dataset from a file, handling potential coordinate name changes."""
  data_tree = output.load_state_file(filename)
  # Handle potential time coordinate name variations
  time = data_tree[output.TIME].to_numpy()

  def get_optional_data(ds, key, grid_type):
    if grid_type.lower() not in ['cell', 'face']:
      raise ValueError(
          f'grid_type for {key} must be either "cell" or "face", got'
          f' {grid_type}'
      )
    if key in ds:
      return ds[key].to_numpy()
    else:
      return (
          np.zeros((len(time), len(ds[output.RHO_CELL_NORM])))
          if grid_type == 'cell'
          else np.zeros((len(time), len(ds[output.RHO_FACE_NORM].to_numpy())))
      )

  def _transform_data(ds: xr.Dataset):
    """Transforms data in-place to the desired units."""
    # TODO(b/414755419)
    ds = ds.copy()

    transformations = {
        output.J_TOTAL: 1e6,  # A/m^2 to MA/m^2
        output.J_OHMIC: 1e6,  # A/m^2 to MA/m^2
        output.J_BOOTSTRAP: 1e6,  # A/m^2 to MA/m^2
        output.J_EXTERNAL: 1e6,  # A/m^2 to MA/m^2
        'j_generic_current': 1e6,  # A/m^2 to MA/m^2
        output.I_BOOTSTRAP: 1e6,  # A to MA
        output.IP_PROFILE: 1e6,  # A to MA
        'j_ecrh': 1e6,  # A/m^2 to MA/m^2
        'p_icrh_i': 1e6,  # W/m^3 to MW/m^3
        'p_icrh_e': 1e6,  # W/m^3 to MW/m^3
        'p_generic_heat_i': 1e6,  # W/m^3 to MW/m^3
        'p_generic_heat_e': 1e6,  # W/m^3 to MW/m^3
        'p_ecrh_e': 1e6,  # W/m^3 to MW/m^3
        'p_alpha_i': 1e6,  # W/m^3 to MW/m^3
        'p_alpha_e': 1e6,  # W/m^3 to MW/m^3
        'p_ohmic_e': 1e6,  # W/m^3 to MW/m^3
        'p_bremsstrahlung_e': 1e6,  # W/m^3 to MW/m^3
        'p_cyclotron_radiation_e': 1e6,  # W/m^3 to MW/m^3
        'p_impurity_radiation_e': 1e6,  # W/m^3 to MW/m^3
        'ei_exchange': 1e6,  # W/m^3 to MW/m^3
        'P_ohmic_e': 1e6,  # W to MW
        'P_aux_total': 1e6,  # W to MW
        'P_alpha_total': 1e6,  # W to MW
        'P_bremsstrahlung_e': 1e6,  # W to MW
        'P_cyclotron_e': 1e6,  # W to MW
        'P_ecrh': 1e6,  # W to MW
        'P_radiation_e': 1e6,  # W to MW
        'I_ecrh': 1e6,  # A to MA
        'I_aux_generic': 1e6,  # A to MA
        'W_thermal_total': 1e6,  # J to MJ
        output.N_E: 1e20,  # m^-3 to 10^{20} m^-3
        output.N_I: 1e20,  # m^-3 to 10^{20} m^-3
        output.N_IMPURITY: 1e20,  # m^-3 to 10^{20} m^-3
    }

    for var_name, scale in transformations.items():
      if var_name in ds:
        ds[var_name] /= scale

    return ds

  data_tree = xr.map_over_datasets(_transform_data, data_tree)
  profiles_dataset = data_tree.children[output.PROFILES].dataset
  scalars_dataset = data_tree.children[output.SCALARS].dataset
  dataset = data_tree.dataset

  return PlotData(
      T_i=profiles_dataset[output.T_I].to_numpy(),
      T_e=profiles_dataset[output.T_E].to_numpy(),
      n_e=profiles_dataset[output.N_E].to_numpy(),
      n_i=profiles_dataset[output.N_I].to_numpy(),
      n_impurity=profiles_dataset[output.N_IMPURITY].to_numpy(),
      Z_impurity=profiles_dataset[output.Z_IMPURITY].to_numpy(),
      psi=profiles_dataset[output.PSI].to_numpy(),
      v_loop=profiles_dataset[output.V_LOOP].to_numpy(),
      j_total=profiles_dataset[output.J_TOTAL].to_numpy(),
      j_ohmic=profiles_dataset[output.J_OHMIC].to_numpy(),
      j_bootstrap=profiles_dataset[output.J_BOOTSTRAP].to_numpy(),
      j_external=profiles_dataset[output.J_EXTERNAL].to_numpy(),
      j_ecrh=get_optional_data(profiles_dataset, 'j_ecrh', 'cell'),
      j_generic_current=get_optional_data(
          profiles_dataset, 'j_generic_current', 'cell'
      ),
      q=profiles_dataset[output.Q].to_numpy(),
      magnetic_shear=profiles_dataset[output.MAGNETIC_SHEAR].to_numpy(),
      chi_turb_i=profiles_dataset[output.CHI_TURB_I].to_numpy(),
      chi_turb_e=profiles_dataset[output.CHI_TURB_E].to_numpy(),
      D_turb_e=profiles_dataset[output.D_TURB_E].to_numpy(),
      V_turb_e=profiles_dataset[output.V_TURB_E].to_numpy(),
      rho_norm=dataset[output.RHO_NORM].to_numpy(),
      rho_cell_norm=dataset[output.RHO_CELL_NORM].to_numpy(),
      rho_face_norm=dataset[output.RHO_FACE_NORM].to_numpy(),
      p_icrh_i=get_optional_data(profiles_dataset, 'p_icrh_i', 'cell'),
      p_icrh_e=get_optional_data(profiles_dataset, 'p_icrh_e', 'cell'),
      p_generic_heat_i=get_optional_data(
          profiles_dataset, 'p_generic_heat_i', 'cell'
      ),
      p_generic_heat_e=get_optional_data(
          profiles_dataset, 'p_generic_heat_e', 'cell'
      ),
      p_ecrh_e=get_optional_data(profiles_dataset, 'p_ecrh_e', 'cell'),
      p_alpha_i=get_optional_data(profiles_dataset, 'p_alpha_i', 'cell'),
      p_alpha_e=get_optional_data(profiles_dataset, 'p_alpha_e', 'cell'),
      p_ohmic_e=get_optional_data(profiles_dataset, 'p_ohmic_e', 'cell'),
      p_bremsstrahlung_e=get_optional_data(
          profiles_dataset, 'p_bremsstrahlung_e', 'cell'
      ),
      p_cyclotron_radiation_e=get_optional_data(
          profiles_dataset, 'p_cyclotron_radiation_e', 'cell'
      ),
      p_impurity_radiation_e=get_optional_data(
          profiles_dataset, 'p_impurity_radiation_e', 'cell'
      ),
      ei_exchange=profiles_dataset[
          'ei_exchange'
      ].to_numpy(),  # ion heating/sink
      Q_fusion=scalars_dataset['Q_fusion'].to_numpy(),  # pylint: disable=invalid-name
      s_gas_puff=get_optional_data(profiles_dataset, 's_gas_puff', 'cell'),
      s_generic_particle=get_optional_data(
          profiles_dataset, 's_generic_particle', 'cell'
      ),
      s_pellet=get_optional_data(profiles_dataset, 's_pellet', 'cell'),
      Ip_profile=profiles_dataset[output.IP_PROFILE].to_numpy()[:, -1],
      I_bootstrap=scalars_dataset[output.I_BOOTSTRAP].to_numpy(),
      I_aux_generic=scalars_dataset['I_aux_generic'].to_numpy(),
      I_ecrh=scalars_dataset['I_ecrh'].to_numpy(),
      P_ohmic_e=scalars_dataset['P_ohmic_e'].to_numpy(),
      P_auxiliary=(
          scalars_dataset['P_aux_total'] - scalars_dataset['P_ohmic_e']
      ).to_numpy(),
      P_alpha_total=scalars_dataset['P_alpha_total'].to_numpy(),
      P_sink=scalars_dataset['P_bremsstrahlung_e'].to_numpy()
      + scalars_dataset['P_radiation_e'].to_numpy()
      + scalars_dataset['P_cyclotron_e'].to_numpy(),
      P_bremsstrahlung_e=scalars_dataset['P_bremsstrahlung_e'].to_numpy(),
      P_radiation_e=scalars_dataset['P_radiation_e'].to_numpy(),
      P_cyclotron_e=scalars_dataset['P_cyclotron_e'].to_numpy(),
      T_e_volume_avg=scalars_dataset['T_e_volume_avg'].to_numpy(),
      T_i_volume_avg=scalars_dataset['T_i_volume_avg'].to_numpy(),
      n_e_volume_avg=scalars_dataset['n_e_volume_avg'].to_numpy(),
      n_i_volume_avg=scalars_dataset['n_i_volume_avg'].to_numpy(),
      W_thermal_total=scalars_dataset['W_thermal_total'].to_numpy(),
      q95=scalars_dataset['q95'].to_numpy(),
      t=time,
  )


def plot_run(
    plot_config: FigureProperties, outfile: str, outfile2: str | None = None
):
  """Plots a single run or comparison of two runs."""
  if not path.exists(outfile):
    raise ValueError(f'File {outfile} does not exist.')
  if outfile2 is not None and not path.exists(outfile2):
    raise ValueError(f'File {outfile2} does not exist.')
  plotdata1 = load_data(outfile)
  plotdata2 = load_data(outfile2) if outfile2 else None

  # Attribute check. Sufficient to check one PlotData object.
  plotdata_attrs = set(
      plotdata1.__dataclass_fields__
  )  # Get PlotData attributes
  for cfg in plot_config.axes:
    for attr in cfg.attrs:
      if attr not in plotdata_attrs:
        raise ValueError(
            f"Attribute '{attr}' in plot_config does not exist in PlotData"
        )

  fig, axes, slider_ax = create_figure(plot_config)

  # Title handling:
  title_lines = [f'(1)={outfile}']
  if outfile2:
    title_lines.append(f'(2)={outfile2}')
  fig.suptitle('\n'.join(title_lines))

  lines1 = get_lines(plot_config, plotdata1, axes)
  lines2 = (
      get_lines(plot_config, plotdata2, axes, comp_plot=True)
      if plotdata2
      else None
  )

  format_plots(plot_config, plotdata1, plotdata2, axes)

  # Only create the slider if needed.
  if plot_config.contains_spatial_plot_type:
    timeslider = create_slider(slider_ax, plotdata1, plotdata2)

    def update(newtime):
      """Update plots with new values following slider manipulation."""
      fig.constrained_layout = False
      _update(newtime, plot_config, plotdata1, lines1, plotdata2, lines2)
      fig.constrained_layout = True
      fig.canvas.draw_idle()

    timeslider.on_changed(update)

  fig.canvas.draw()
  plt.show()


def _update(
    newtime,
    plot_config: FigureProperties,
    plotdata1: PlotData,
    lines1: Sequence[matplotlib.lines.Line2D],
    plotdata2: PlotData | None = None,
    lines2: Sequence[matplotlib.lines.Line2D] | None = None,
):
  """Update plots with new values following slider manipulation."""

  def update_lines(plotdata, lines):
    idx = np.abs(plotdata.t - newtime).argmin()
    line_idx = 0
    for cfg in plot_config.axes:  # Iterate through axes based on plot_config
      if cfg.plot_type == PlotType.TIME_SERIES:
        continue  # Time series plots do not need to be updated
      for attr in cfg.attrs:  # Update all lines in current subplot.
        data = getattr(plotdata, attr)
        if cfg.suppress_zero_values and np.all(data == 0):
          continue
        lines[line_idx].set_ydata(data[idx, :])
        line_idx += 1

  update_lines(plotdata1, lines1)
  if plotdata2 and lines2:
    update_lines(plotdata2, lines2)


def create_slider(
    ax: matplotlib.axes.Axes,
    plotdata1: PlotData,
    plotdata2: PlotData | None = None,
) -> widgets.Slider:
  """Create a slider tool for the plot."""
  tmin = (
      min(plotdata1.t)
      if plotdata2 is None
      else min(min(plotdata1.t), min(plotdata2.t))
  )
  tmax = (
      max(plotdata1.t)
      if plotdata2 is None
      else max(max(plotdata1.t), max(plotdata2.t))
  )

  dt = (
      min(np.diff(plotdata1.t))
      if plotdata2 is None
      else min(min(np.diff(plotdata1.t)), min(np.diff(plotdata2.t)))
  )

  return widgets.Slider(
      ax,
      'Time [s]',
      tmin,
      tmax,
      valinit=tmin,
      valstep=dt,
  )


def format_plots(
    plot_config: FigureProperties,
    plotdata1: PlotData,
    plotdata2: PlotData | None,
    axes: List[Any],
):
  """Sets up plot formatting."""

  # Set default legend fontsize for legends
  matplotlib.rc('legend', fontsize=plot_config.default_legend_fontsize)

  def get_limit(plotdata, attrs, percentile, include_first_timepoint):
    """Gets the limit for a set of attributes based a histogram percentile."""
    if include_first_timepoint:
      values = np.concatenate(
          [getattr(plotdata, attr).flatten() for attr in attrs]
      )
    else:
      values = np.concatenate(
          [getattr(plotdata, attr)[1:, :].flatten() for attr in attrs]
      )
    return np.percentile(values, percentile)

  for ax, cfg in zip(axes, plot_config.axes):
    if cfg.plot_type == PlotType.SPATIAL:
      ax.set_xlabel('Normalized radius')
    elif cfg.plot_type == PlotType.TIME_SERIES:
      ax.set_xlabel('Time [s]')
    else:
      raise ValueError(f'Unknown plot type: {cfg.plot_type}')
    ax.set_ylabel(cfg.ylabel)

    # Get limits for y-axis based on percentile values.
    # 0.0 or 100.0 are special cases for simple min/max values.
    ymin = get_limit(
        plotdata1, cfg.attrs, cfg.lower_percentile, cfg.include_first_timepoint
    )
    ymax = get_limit(
        plotdata1, cfg.attrs, cfg.upper_percentile, cfg.include_first_timepoint
    )

    if plotdata2:
      ymin = min(
          ymin,
          get_limit(
              plotdata2,
              cfg.attrs,
              cfg.lower_percentile,
              cfg.include_first_timepoint,
          ),
      )
      ymax = max(
          ymax,
          get_limit(
              plotdata2,
              cfg.attrs,
              cfg.upper_percentile,
              cfg.include_first_timepoint,
          ),
      )

    lower_bound = ymin / 1.05 if ymin > 0 else ymin * 1.05

    # Guard against empty data
    if ymax != 0 or ymin != 0:  # Check for meaningful data range
      if cfg.ylim_min_zero:
        ax.set_ylim([min(lower_bound, 0), ymax * 1.05])
      else:
        ax.set_ylim([lower_bound, ymax * 1.05])

      ax.legend(fontsize=cfg.legend_fontsize)


def get_rho(
    plotdata: PlotData,
    data_attr: str,
) -> np.ndarray:
  """Gets the correct rho coordinate for the data."""
  datalen = len(getattr(plotdata, data_attr)[0, :])
  if datalen == len(plotdata.rho_cell_norm):
    return plotdata.rho_cell_norm
  elif datalen == len(plotdata.rho_face_norm):
    return plotdata.rho_face_norm
  elif datalen == len(plotdata.rho_norm):
    return plotdata.rho_norm
  else:
    raise ValueError(
        f'Data {datalen} does not coincide with either the cell or face grids.'
    )


def get_lines(
    plot_config: FigureProperties,
    plotdata: PlotData,
    axes: List[Any],
    comp_plot: bool = False,
):
  """Gets lines for all plots."""
  lines = []
  # If comparison, first lines labeled (1) and solid, second set (2) and dashed.
  suffix = f' ({1 if not comp_plot else 2})'
  dashed = '--' if comp_plot else ''

  for ax, cfg in zip(axes, plot_config.axes):
    line_idx = 0  # Reset color selection cycling for each plot.
    if cfg.plot_type == PlotType.SPATIAL:
      for attr, label in zip(cfg.attrs, cfg.labels):
        data = getattr(plotdata, attr)
        if cfg.suppress_zero_values and np.all(data == 0):
          continue
        rho = get_rho(plotdata, attr)
        (line,) = ax.plot(
            rho,
            data[0, :],  # Plot data at time zero
            plot_config.colors[line_idx % len(plot_config.colors)] + dashed,
            label=f'{label}{suffix}',
        )
        lines.append(line)
        line_idx += 1
    elif cfg.plot_type == PlotType.TIME_SERIES:
      for attr, label in zip(cfg.attrs, cfg.labels):
        data = getattr(plotdata, attr)
        if cfg.suppress_zero_values and np.all(data == 0):
          continue
        # No need to return a line since this will not need to be updated.
        _ = ax.plot(
            plotdata.t,
            data,  # Plot entire time series
            plot_config.colors[line_idx % len(plot_config.colors)] + dashed,
            label=f'{label}{suffix}',
        )
        line_idx += 1
    else:
      raise ValueError(f'Unknown plot type: {cfg.plot_type}')

  return lines


def create_figure(plot_config: FigureProperties):
  """Creates the figure and axes."""
  rows = plot_config.rows
  cols = plot_config.cols
  matplotlib.rc('xtick', labelsize=plot_config.tick_fontsize)
  matplotlib.rc('ytick', labelsize=plot_config.tick_fontsize)
  matplotlib.rc('axes', labelsize=plot_config.axes_fontsize)
  matplotlib.rc('figure', titlesize=plot_config.title_fontsize)
  fig = plt.figure(
      figsize=(
          cols * plot_config.figure_size_factor,
          rows * plot_config.figure_size_factor,
      ),
      constrained_layout=True,
  )
  # Create the GridSpec - Adjust height ratios to include the slider
  # in the plot, only if a slider is required:
  if plot_config.contains_spatial_plot_type:
    # Add an extra smaller is a spatial plottypeider
    height_ratios = [1] * rows + [0.2]
    gs = gridspec.GridSpec(
        rows + 1, cols, figure=fig, height_ratios=height_ratios
    )
    # slider spans all columns
    slider_ax = fig.add_subplot(gs[rows, :])
  else:
    gs = gridspec.GridSpec(rows, cols, figure=fig)
    slider_ax = None

  axes = []
  for i in range(rows * cols):
    row = i // cols
    col = i % cols
    axes.append(fig.add_subplot(gs[row, col]))  # Add subplots to the grid
  return fig, axes, slider_ax



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains the main programmatic entry point for running a TORAX simulation.

The intended use is
```
torax_config = torax.ToraxConfig.from_dict(config_dict)
sim_outputs = torax.run_simulation(torax_config)

# Update the config to run a new simulation with different parameters.
torax_config.update(updated_fields)
new_sim_outputs = torax.run_simulation(torax_config)
```
"""

from torax._src.config import build_runtime_params
from torax._src.orchestration import initial_state as initial_state_lib
from torax._src.orchestration import run_loop
from torax._src.orchestration import step_function
from torax._src.output_tools import output
from torax._src.sources import source_models as source_models_lib
from torax._src.torax_pydantic import model_config
import xarray as xr


def run_simulation(
    torax_config: model_config.ToraxConfig,
    log_timestep_info: bool = False,
    progress_bar: bool = True,
) -> tuple[xr.DataTree, output.StateHistory]:
  """Runs a TORAX simulation using the config and returns the outputs.

  Args:
    torax_config: The TORAX config to use for the simulation.
    log_timestep_info: Whether to log the timestep information.
    progress_bar: Whether to show a progress bar.

  Returns:
    A tuple of the simulation outputs in the form of a DataTree and the state
    history which is intended for helpful use with debugging as it contains
    the `CoreProfiles`, `CoreTransport`, `CoreSources`, `Geometry`, and
    `PostProcessedOutputs` dataclasses for each step of the simulation.
  """
  # TODO(b/384767453): Remove the need for the step_fn and solver to take the
  # transport model and pedestal model.
  transport_model = torax_config.transport.build_transport_model()
  pedestal_model = torax_config.pedestal.build_pedestal_model()

  geometry_provider = torax_config.geometry.build_provider
  source_models = source_models_lib.SourceModels(
      torax_config.sources, neoclassical=torax_config.neoclassical
  )

  static_runtime_params_slice = (
      build_runtime_params.build_static_params_from_config(torax_config)
  )

  solver = torax_config.solver.build_solver(
      static_runtime_params_slice=static_runtime_params_slice,
      transport_model=transport_model,
      source_models=source_models,
      pedestal_model=pedestal_model,
  )

  mhd_models = torax_config.mhd.build_mhd_models(
      static_runtime_params_slice=static_runtime_params_slice,
      transport_model=transport_model,
      source_models=source_models,
      pedestal_model=pedestal_model,
  )

  step_fn = step_function.SimulationStepFn(
      solver=solver,
      time_step_calculator=torax_config.time_step_calculator.time_step_calculator,
      transport_model=transport_model,
      pedestal_model=pedestal_model,
      mhd_models=mhd_models,
  )

  dynamic_runtime_params_slice_provider = (
      build_runtime_params.DynamicRuntimeParamsSliceProvider.from_config(
          torax_config
      )
  )

  if torax_config.restart and torax_config.restart.do_restart:
    initial_state, post_processed_outputs = (
        initial_state_lib.get_initial_state_and_post_processed_outputs_from_file(
            t_initial=torax_config.numerics.t_initial,
            file_restart=torax_config.restart,
            static_runtime_params_slice=static_runtime_params_slice,
            dynamic_runtime_params_slice_provider=dynamic_runtime_params_slice_provider,
            geometry_provider=geometry_provider,
            step_fn=step_fn,
        )
    )
    restart_case = True
  else:
    initial_state, post_processed_outputs = (
        initial_state_lib.get_initial_state_and_post_processed_outputs(
            t=torax_config.numerics.t_initial,
            static_runtime_params_slice=static_runtime_params_slice,
            dynamic_runtime_params_slice_provider=dynamic_runtime_params_slice_provider,
            geometry_provider=geometry_provider,
            step_fn=step_fn,
        )
    )
    restart_case = False

  state_history, post_processed_outputs_history, sim_error = run_loop.run_loop(
      static_runtime_params_slice=static_runtime_params_slice,
      dynamic_runtime_params_slice_provider=dynamic_runtime_params_slice_provider,
      geometry_provider=geometry_provider,
      initial_state=initial_state,
      initial_post_processed_outputs=post_processed_outputs,
      restart_case=restart_case,
      step_fn=step_fn,
      log_timestep_info=log_timestep_info,
      progress_bar=progress_bar,
  )
  state_history = output.StateHistory(
      state_history=state_history,
      post_processed_outputs_history=post_processed_outputs_history,
      sim_error=sim_error,
      torax_config=torax_config,
  )

  return (
      state_history.simulation_output_to_xr(torax_config.restart),
      state_history,
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Calculations related to derived quantities from poloidal flux (psi).

Functions:
    - calc_q: Calculates the q-profile (q).
    - calc_j_total: Calculate flux-surface-averaged toroidal current density.
    - calc_s: Calculates magnetic shear (s).
    - calc_s_rmid: Calculates magnetic shear (s), using midplane r as radial
      coordinate.
    - calc_Wpol: Calculates total magnetic energy (Wpol).
    - calc_li3: Calculates normalized internal inductance li3 (ITER convention).
    - calc_q95: Calculates the q-profile at 95% of the normalized poloidal flux.
    - calculate_psi_grad_constraint_from_Ip: Calculates the gradient
      constraint on the poloidal flux (psi) from Ip.
    - _calc_bpol2: Calculates square of poloidal field (Bp).
"""
import chex
import jax
from jax import numpy as jnp
from torax._src import array_typing
from torax._src import constants
from torax._src import jax_utils
from torax._src.fvm import cell_variable
from torax._src.fvm import convection_terms
from torax._src.fvm import diffusion_terms
from torax._src.geometry import geometry

_trapz = jax.scipy.integrate.trapezoid

# pylint: disable=invalid-name


def calc_q_face(
    geo: geometry.Geometry,
    psi: cell_variable.CellVariable,
) -> chex.Array:
  """Calculates the q-profile on the face grid given poloidal flux (psi)."""
  # iota is standard terminology for 1/q
  inv_iota = jnp.abs(
      (2 * geo.Phi_b * geo.rho_face_norm[1:]) / psi.face_grad()[1:]
  )

  # Use L'Hôpital's rule to calculate iota on-axis, with psi_face_grad()[0]=0.
  inv_iota0 = jnp.expand_dims(
      jnp.abs((2 * geo.Phi_b * geo.drho_norm) / psi.face_grad()[1]), 0
  )

  q_face = jnp.concatenate([inv_iota0, inv_iota])
  return q_face * geo.q_correction_factor


def calc_j_total(
    geo: geometry.Geometry,
    psi: cell_variable.CellVariable,
) -> tuple[chex.Array, chex.Array, chex.Array]:
  """Calculate flux-surface-averaged toroidal current density from poloidal flux.

  Calculation based on j_total = dI/dS

  Args:
    geo: Torus geometry.
    psi: Poloidal flux.

  Returns:
    j_total: total current density [A/m2] on cell grid
    j_total_face: total current density [A/m2] on face grid
    Ip_profile_face: cumulative total plasma current profile [A] on face grid
  """

  # pylint: disable=invalid-name
  Ip_profile_face = (
      psi.face_grad()
      * geo.g2g3_over_rhon_face
      * geo.F_face
      / geo.Phi_b
      / (16 * jnp.pi**3 * constants.CONSTANTS.mu0)
  )

  Ip_profile = (
      psi.grad()
      * geo.g2g3_over_rhon
      * geo.F
      / geo.Phi_b
      / (16 * jnp.pi**3 * constants.CONSTANTS.mu0)
  )

  dI_drhon_face = jnp.gradient(Ip_profile_face, geo.rho_face_norm)
  dI_drhon = jnp.gradient(Ip_profile, geo.rho_norm)

  j_total_bulk = dI_drhon[1:] / geo.spr[1:]
  j_total_face_bulk = dI_drhon_face[1:] / geo.spr_face[1:]

  # Extrapolate the axis term from the bulk term due to strong sensitivities
  # of near-axis numerical derivatives. Set zero boundary condition on-axis
  j_total_axis = j_total_bulk[0] - (j_total_bulk[1] - j_total_bulk[0])

  j_total = jnp.concatenate([jnp.array([j_total_axis]), j_total_bulk])
  j_total_face = jnp.concatenate([jnp.array([j_total_axis]), j_total_face_bulk])

  return j_total, j_total_face, Ip_profile_face


def calc_s_face(
    geo: geometry.Geometry, psi: cell_variable.CellVariable
) -> jax.Array:
  """Calculates magnetic shear on the face grid from poloidal flux (psi)."""

  # iota (1/q) should have a /2*Phib but we drop it since will cancel out in
  # the s calculation.
  iota_scaled = jnp.abs((psi.face_grad()[1:] / geo.rho_face_norm[1:]))

  # on-axis iota_scaled from L'Hôpital's rule = dpsi_face_grad / drho_norm
  # Using expand_dims to make it compatible with jnp.concatenate
  iota_scaled0 = jnp.expand_dims(
      jnp.abs(psi.face_grad()[1] / geo.drho_norm), axis=0
  )

  iota_scaled = jnp.concatenate([iota_scaled0, iota_scaled])

  s_face = (
      -geo.rho_face_norm
      * jnp.gradient(iota_scaled, geo.rho_face_norm)
      / iota_scaled
  )

  return s_face


def calc_s_rmid(
    geo: geometry.Geometry, psi: cell_variable.CellVariable
) -> jax.Array:
  """Calculates magnetic shear (s) from poloidal flux (psi).

  Version taking the derivative of iota with respect to the midplane r,
  in line with expectations from circular-derived models like QuaLiKiz.

  Args:
    geo: Torus geometry.
    psi: Poloidal flux.

  Returns:
    s_face: Magnetic shear, on the face grid.
  """

  # iota (1/q) should have a /2*Phib but we drop it since will cancel out in
  # the s calculation.
  iota_scaled = jnp.abs((psi.face_grad()[1:] / geo.rho_face_norm[1:]))

  # on-axis iota_scaled from L'Hôpital's rule = dpsi_face_grad / drho_norm
  # Using expand_dims to make it compatible with jnp.concatenate
  iota_scaled0 = jnp.expand_dims(
      jnp.abs(psi.face_grad()[1] / geo.drho_norm), axis=0
  )

  iota_scaled = jnp.concatenate([iota_scaled0, iota_scaled])

  rmid_face = (geo.R_out_face - geo.R_in_face) * 0.5

  s_face = -rmid_face * jnp.gradient(iota_scaled, rmid_face) / iota_scaled

  return s_face


def _calc_bpol2(
    geo: geometry.Geometry, psi: cell_variable.CellVariable
) -> jax.Array:
  r"""Calculates square of poloidal field (Bp) from poloidal flux (psi).

  An identity for the poloidal magnetic field is:
  :math:`B_p = 1/R \partial \psi / \partial \rho (\nabla \rho \times e_\phi)`

  Where :math:`e_\phi` is the unit vector pointing in the toroidal direction.

  Args:
    geo: Torus geometry.
    psi: Poloidal flux.

  Returns:
    bpol2_face: Square of poloidal magnetic field, on the face grid.
  """
  bpol2_bulk = (
      (psi.face_grad()[1:] / (2 * jnp.pi)) ** 2
      * geo.g2_face[1:]
      / geo.vpr_face[1:] ** 2
  )
  bpol2_axis = jnp.array([0.0], dtype=jax_utils.get_dtype())
  bpol2_face = jnp.concatenate([bpol2_axis, bpol2_bulk])
  return bpol2_face


def calc_Wpol(
    geo: geometry.Geometry, psi: cell_variable.CellVariable
) -> jax.Array:
  """Calculates total magnetic energy (Wpol) from poloidal flux (psi)."""
  bpol2 = _calc_bpol2(geo, psi)
  Wpol = _trapz(bpol2 * geo.vpr_face, geo.rho_face_norm) / (
      2 * constants.CONSTANTS.mu0
  )
  return Wpol


def calc_li3(
    R_major: jax.Array,
    Wpol: jax.Array,
    Ip_total: jax.Array,
) -> jax.Array:
  """Calculates li3 based on a formulation using Wpol.

  Normalized internal inductance is defined as:
  li = <Bpol^2>_V / <Bpol^2>_LCFS where <>_V is a volume average and <>_LCFS is
  the average at the last closed flux surface.

  We use the ITER convention for normalized internal inductance defined as:
  li3 = 2*V*<Bpol^2>_V / (mu0^2 Ip^2*R_major) = 4 * Wpol / (mu0 Ip^2*R_major)

  Ip (total plasma current) enters through the integral form of Ampere's law.
  Since Wpol also corresponds to a volume integral of the poloidal field, we
  can define li3 with respect to Wpol.

  Args:
    R_major: Major radius.
    Wpol: Total magnetic energy.
    Ip_total: Total plasma current.

  Returns:
    li3: Normalized internal inductance, ITER convention.
  """
  return 4 * Wpol / (constants.CONSTANTS.mu0 * Ip_total**2 * R_major)


def calc_q95(
    psi_norm_face: array_typing.ArrayFloat,
    q_face: array_typing.ArrayFloat,
) -> array_typing.ScalarFloat:
  """Calculates q95 from the q profile and the normalized poloidal flux.

  Args:
    psi_norm_face: normalized poloidal flux
    q_face: safety factor on the face grid

  Returns:
    q95: q at 95% of the normalized poloidal flux
  """
  q95 = jnp.interp(0.95, psi_norm_face, q_face)

  return q95


def calculate_psi_grad_constraint_from_Ip(
    Ip: array_typing.ScalarFloat,
    geo: geometry.Geometry,
) -> jax.Array:
  """Calculates the gradient constraint on the poloidal flux (psi) from Ip."""
  return (
      Ip
      * (16 * jnp.pi**3 * constants.CONSTANTS.mu0 * geo.Phi_b)
      / (geo.g2g3_over_rhon_face[-1] * geo.F_face[-1])
  )


def calculate_psidot_from_psi_sources(
    *,
    psi_sources: array_typing.ArrayFloat,
    sigma: array_typing.ArrayFloat,
    sigma_face: array_typing.ArrayFloat,
    resistivity_multiplier: float,
    psi: cell_variable.CellVariable,
    geo: geometry.Geometry,
) -> jax.Array:
  """Calculates psidot (loop voltage) from the sum of the psi sources."""

  # Calculate transient term
  consts = constants.CONSTANTS
  toc_psi = (
      1.0
      / resistivity_multiplier
      * geo.rho_norm
      * sigma
      * consts.mu0
      * 16
      * jnp.pi**2
      * geo.Phi_b**2
      / geo.F**2
  )
  # Calculate diffusion term coefficient
  d_face_psi = geo.g2g3_over_rhon_face
  # Add phibdot terms to poloidal flux convection
  v_face_psi = (
      -8.0
      * jnp.pi**2
      * consts.mu0
      * geo.Phi_b_dot
      * geo.Phi_b
      * sigma_face
      * geo.rho_face_norm**2
      / geo.F_face**2
  )

  # Add effective phibdot poloidal flux source term
  ddrnorm_sigma_rnorm2_over_f2 = jnp.gradient(
      sigma * geo.rho_norm**2 / geo.F**2, geo.rho_norm
  )

  psi_sources += (
      -8.0
      * jnp.pi**2
      * consts.mu0
      * geo.Phi_b_dot
      * geo.Phi_b
      * ddrnorm_sigma_rnorm2_over_f2
  )

  diffusion_mat, diffusion_vec = diffusion_terms.make_diffusion_terms(
      d_face_psi, psi
  )

  # Set the psi convection term for psidot used in ohmic power, always with
  # the default 'ghost' mode. Impact of different modes would mildly impact
  # Ohmic power at the LCFS which has negligible impact on simulations.
  # Allowing it to be configurable introduces more complexity in the code by
  # needing to pass in the mode from the static_runtime_params across multiple
  # functions.
  conv_mat, conv_vec = convection_terms.make_convection_terms(
      v_face_psi, d_face_psi, psi
  )

  c_mat = diffusion_mat + conv_mat
  c = diffusion_vec + conv_vec

  c += psi_sources

  psidot = (jnp.dot(c_mat, psi.value) + c) / toc_psi

  return psidot



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sauter model implementation - to be moved to neoclassical."""
import chex
from jax import numpy as jnp
from torax._src import constants
from torax._src import jax_utils
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.sources import source_profiles


@jax_utils.jit
def calc_sauter_model(
    *,
    bootstrap_multiplier: float,
    density_reference: float,
    Z_eff_face: chex.Array,
    Z_i_face: chex.Array,
    n_e: cell_variable.CellVariable,
    n_i: cell_variable.CellVariable,
    T_e: cell_variable.CellVariable,
    T_i: cell_variable.CellVariable,
    psi: cell_variable.CellVariable,
    q_face: chex.Array,
    geo: geometry.Geometry,
) -> source_profiles.BootstrapCurrentProfile:
  """Calculates sigmaneo, j_bootstrap, and I_bootstrap."""
  # pylint: disable=invalid-name

  # Formulas from Sauter PoP 1999. Future work can include Redl PoP 2021
  # corrections.

  true_n_e_face = n_e.face_value() * density_reference
  true_n_i_face = n_i.face_value() * density_reference

  # # local r/R0 on face grid
  epsilon = (geo.R_out_face - geo.R_in_face) / (geo.R_out_face + geo.R_in_face)
  epseff = (
      0.67 * (1.0 - 1.4 * jnp.abs(geo.delta_face) * geo.delta_face) * epsilon
  )
  aa = (1.0 - epsilon) / (1.0 + epsilon)
  ftrap = 1.0 - jnp.sqrt(aa) * (1.0 - epseff) / (1.0 + 2.0 * jnp.sqrt(epseff))

  # Spitzer conductivity
  NZ = 0.58 + 0.74 / (0.76 + Z_eff_face)
  lnLame = (
      31.3 - 0.5 * jnp.log(true_n_e_face) + jnp.log(T_e.face_value() * 1e3)
  )
  lnLami = (
      30
      - 3 * jnp.log(Z_i_face)
      - 0.5 * jnp.log(true_n_i_face)
      + 1.5 * jnp.log(T_i.face_value() * 1e3)
  )

  sigsptz = (
      1.9012e04 * (T_e.face_value() * 1e3) ** 1.5 / Z_eff_face / NZ / lnLame
  )

  nuestar = (
      6.921e-18
      * q_face
      * geo.R_major
      * true_n_e_face
      * Z_eff_face
      * lnLame
      / (
          ((T_e.face_value() * 1e3) ** 2)
          * (epsilon + constants.CONSTANTS.eps) ** 1.5
      )
  )
  nuistar = (
      4.9e-18
      * q_face
      * geo.R_major
      * true_n_i_face
      * Z_eff_face**4
      * lnLami
      / (
          ((T_i.face_value() * 1e3) ** 2)
          * (epsilon + constants.CONSTANTS.eps) ** 1.5
      )
  )

  # Neoclassical correction to spitzer conductivity
  ft33 = ftrap / (
      1.0
      + (0.55 - 0.1 * ftrap) * jnp.sqrt(nuestar)
      + 0.45 * (1.0 - ftrap) * nuestar / (Z_eff_face**1.5)
  )
  signeo_face = 1.0 - ft33 * (
      1.0
      + 0.36 / Z_eff_face
      - ft33 * (0.59 / Z_eff_face - 0.23 / Z_eff_face * ft33)
  )
  sigmaneo = sigsptz * signeo_face

  # Calculate terms needed for bootstrap current
  denom = (
      1.0
      + (1 - 0.1 * ftrap) * jnp.sqrt(nuestar)
      + 0.5 * (1.0 - ftrap) * nuestar / Z_eff_face
  )
  ft31 = ftrap / denom
  ft32ee = ftrap / (
      1
      + 0.26 * (1 - ftrap) * jnp.sqrt(nuestar)
      + 0.18 * (1 - 0.37 * ftrap) * nuestar / jnp.sqrt(Z_eff_face)
  )
  ft32ei = ftrap / (
      1
      + (1 + 0.6 * ftrap) * jnp.sqrt(nuestar)
      + 0.85 * (1 - 0.37 * ftrap) * nuestar * (1 + Z_eff_face)
  )
  ft34 = ftrap / (
      1.0
      + (1 - 0.1 * ftrap) * jnp.sqrt(nuestar)
      + 0.5 * (1.0 - 0.5 * ftrap) * nuestar / Z_eff_face
  )

  F32ee = (
      (0.05 + 0.62 * Z_eff_face)
      / (Z_eff_face * (1 + 0.44 * Z_eff_face))
      * (ft32ee - ft32ee**4)
      + 1
      / (1 + 0.22 * Z_eff_face)
      * (ft32ee**2 - ft32ee**4 - 1.2 * (ft32ee**3 - ft32ee**4))
      + 1.2 / (1 + 0.5 * Z_eff_face) * ft32ee**4
  )

  F32ei = (
      -(0.56 + 1.93 * Z_eff_face)
      / (Z_eff_face * (1 + 0.44 * Z_eff_face))
      * (ft32ei - ft32ei**4)
      + 4.95
      / (1 + 2.48 * Z_eff_face)
      * (ft32ei**2 - ft32ei**4 - 0.55 * (ft32ei**3 - ft32ei**4))
      - 1.2 / (1 + 0.5 * Z_eff_face) * ft32ei**4
  )

  term_0 = (1 + 1.4 / (Z_eff_face + 1)) * ft31
  term_1 = -1.9 / (Z_eff_face + 1) * ft31**2
  term_2 = 0.3 / (Z_eff_face + 1) * ft31**3
  term_3 = 0.2 / (Z_eff_face + 1) * ft31**4
  L31 = term_0 + term_1 + term_2 + term_3

  L32 = F32ee + F32ei

  L34 = (
      (1 + 1.4 / (Z_eff_face + 1)) * ft34
      - 1.9 / (Z_eff_face + 1) * ft34**2
      + 0.3 / (Z_eff_face + 1) * ft34**3
      + 0.2 / (Z_eff_face + 1) * ft34**4
  )

  alpha0 = -1.17 * (1 - ftrap) / (1 - 0.22 * ftrap - 0.19 * ftrap**2)
  alpha = (
      alpha0
      + 0.25
      * (1 - ftrap**2)
      * jnp.sqrt(nuistar)
      / (1 + 0.5 * jnp.sqrt(nuistar))
      + 0.315 * nuistar**2 * ftrap**6
  ) / (1 + 0.15 * nuistar**2 * ftrap**6)

  # calculate bootstrap current
  prefactor = -geo.F_face * bootstrap_multiplier * 2 * jnp.pi / geo.B_0

  pe = true_n_e_face * (T_e.face_value()) * 1e3 * 1.6e-19
  pi = true_n_i_face * (T_i.face_value()) * 1e3 * 1.6e-19

  dpsi_drnorm = psi.face_grad()
  dlnne_drnorm = n_e.face_grad() / n_e.face_value()
  dlnni_drnorm = n_i.face_grad() / n_i.face_value()
  dlnte_drnorm = T_e.face_grad() / T_e.face_value()
  dlnti_drnorm = T_i.face_grad() / T_i.face_value()

  global_coeff = prefactor[1:] / dpsi_drnorm[1:]
  global_coeff = jnp.concatenate([jnp.zeros(1), global_coeff])

  necoeff = L31 * pe
  nicoeff = L31 * pi
  tecoeff = (L31 + L32) * pe
  ticoeff = (L31 + alpha * L34) * pi

  j_bootstrap_face = global_coeff * (
      necoeff * dlnne_drnorm
      + nicoeff * dlnni_drnorm
      + tecoeff * dlnte_drnorm
      + ticoeff * dlnti_drnorm
  )

  #  j_bootstrap_face = jnp.concatenate([jnp.zeros(1), j_bootstrap_face])
  j_bootstrap_face = jnp.array(j_bootstrap_face)
  j_bootstrap = geometry.face_to_cell(j_bootstrap_face)
  sigmaneo_cell = geometry.face_to_cell(sigmaneo)

  return source_profiles.BootstrapCurrentProfile(
      sigma=sigmaneo_cell,
      sigma_face=sigmaneo,
      j_bootstrap=j_bootstrap,
      j_bootstrap_face=j_bootstrap_face,
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sawtooth model."""

import dataclasses
import jax
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.mhd.sawtooth import redistribution_base
from torax._src.mhd.sawtooth import trigger_base
from torax._src.neoclassical.conductivity import base as base_conductivity
from torax._src.pedestal_model import pedestal_model as pedestal_model_lib
from torax._src.solver import solver
from torax._src.sources import source_models as source_models_lib
from torax._src.sources import source_profile_builders
from torax._src.sources import source_profiles as source_profiles_lib
from torax._src.transport_model import transport_model as transport_model_lib


# TODO(b/414537757). Sawtooth extensions.
# a. Full and incomplete Kadomtsev redistribution model.
# b. Porcelli model with free parameters and fast ion sensitivities.
# c. "Smooth" version that can work with forward-sensitivity-analysis and
#    stationary-state applications without the need for averaging.
class SawtoothModel(solver.Solver):
  """Sawtooth trigger and redistribution, and carries out sawtooth step."""

  def __init__(
      self,
      static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
      trigger_model: trigger_base.TriggerModel,
      redistribution_model: redistribution_base.RedistributionModel,
      transport_model: transport_model_lib.TransportModel,
      pedestal_model: pedestal_model_lib.PedestalModel,
      source_models: source_models_lib.SourceModels,
  ):
    super().__init__(
        static_runtime_params_slice=static_runtime_params_slice,
        transport_model=transport_model,
        source_models=source_models,
        pedestal_model=pedestal_model,
    )
    self._trigger_model = trigger_model
    self._redistribution_model = redistribution_model

  def _x_new(
      self,
      dt: jax.Array,
      static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
      dynamic_runtime_params_slice_t: runtime_params_slice.DynamicRuntimeParamsSlice,
      dynamic_runtime_params_slice_t_plus_dt: runtime_params_slice.DynamicRuntimeParamsSlice,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
      core_profiles_t_plus_dt: state.CoreProfiles,
      core_sources_t: source_profiles_lib.SourceProfiles,
      core_transport_t: state.CoreTransport,
      explicit_source_profiles: source_profiles_lib.SourceProfiles,
      evolving_names: tuple[str, ...],
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      source_profiles_lib.SourceProfiles,
      base_conductivity.Conductivity,
      state.CoreTransport,
      state.SolverNumericOutputs,
  ]:
    """Applies the sawtooth model and outputs new state attributes if triggered.

    If the trigger model indicates a crash has been triggered, an
    instantaneous redistribution model is applied. New state attributes
    following a short (configurable) dt are returned. Beyond the sawtooth
    redistribution, core_profiles are further updated by: the psidot assumed at
    time t; the new boundary conditions; sources, transport, geo, and pedestal
    outputs consistent with the new core_profiles at time t_plus_crash_dt.

    Args:
      dt: Sawtooth step duration.
      static_runtime_params_slice: Static runtime parameters.
      dynamic_runtime_params_slice_t: Dynamic runtime parameters at time t.
      dynamic_runtime_params_slice_t_plus_dt: Dynamic runtime parameters at time
        t + crash_dt.
      geo_t: Geometry at time t.
      geo_t_plus_dt: Geometry at time t + crash_dt.
      core_profiles_t: Core profiles at time t.
      core_profiles_t_plus_dt: Core profiles containing boundary conditions and
        prescribed profiles at time t + crash_dt.
      core_sources_t: Source profiles at time t.
      core_transport_t: Transport coefficients at time t.
      explicit_source_profiles: Explicit source profiles at time t.
      evolving_names: Names of evolving variables.

    Returns:
      Updated tuple of evolving CellVariables from CoreProfiles
      Source profiles consistent with redistributed state.
      Conductivity consistent with redistributed state.
      Transport coefficients consistent with redistributed state.
      SolverNumericOutputs indicating a sawtooth crash.
    """

    trigger_sawtooth, rho_norm_q1 = self._trigger_model(
        static_runtime_params_slice,
        dynamic_runtime_params_slice_t,
        geo_t,
        core_profiles_t,
    )

    def _redistribute_state() -> tuple[
        tuple[cell_variable.CellVariable, ...],
        source_profiles_lib.SourceProfiles,
        base_conductivity.Conductivity,
        state.CoreTransport,
        state.SolverNumericOutputs,
    ]:

      redistributed_core_profiles = self._redistribution_model(
          rho_norm_q1,
          static_runtime_params_slice,
          dynamic_runtime_params_slice_t,
          geo_t,
          core_profiles_t,
      )

      # Evolve the psi profile over the sawtooth time.
      # Redistribution maintains the same psi boundary condition. However,
      # over the course of the sawtooth time, the central solenoid must still
      # modify the psi profile. Since we don't calculate the psi PDE here, we
      # assume that for the short sawtooth time we can use the psidot from the
      # beginning of the step interval. This updates the bulk values. Later, the
      # boundary conditions are also updated at time t_plus_dt when
      # using `updaters.update_all_core_profiles_after_step`.
      evolved_psi_redistributed_value = (
          redistributed_core_profiles.psi.value
          + core_profiles_t.psidot.value * dt
      )
      evolved_core_profiles = dataclasses.replace(
          redistributed_core_profiles,
          psi=dataclasses.replace(
              redistributed_core_profiles.psi,
              value=evolved_psi_redistributed_value,
          ),
      )

      conductivity_post_step = (
          self.source_models.conductivity.calculate_conductivity(
              dynamic_runtime_params_slice_t_plus_dt,
              geo_t_plus_dt,
              redistributed_core_profiles,
          )
      )
      core_sources_post_step = source_profile_builders.get_all_source_profiles(
          static_runtime_params_slice,
          dynamic_runtime_params_slice_t_plus_dt,
          geo_t_plus_dt,
          core_profiles=evolved_core_profiles,
          source_models=self.source_models,
          conductivity=conductivity_post_step,
      )

      x_post_step = tuple(
          [getattr(evolved_core_profiles, name) for name in evolving_names]
      )

      pedestal_model_output = self.pedestal_model(
          dynamic_runtime_params_slice_t_plus_dt,
          geo_t_plus_dt,
          redistributed_core_profiles,
      )

      core_transport_post_step = self.transport_model(
          dynamic_runtime_params_slice_t_plus_dt,
          geo_t_plus_dt,
          redistributed_core_profiles,
          pedestal_model_output,
      )

      solver_numeric_outputs_post_step = state.SolverNumericOutputs(
          sawtooth_crash=True
      )

      return (
          x_post_step,
          core_sources_post_step,
          conductivity_post_step,
          core_transport_post_step,
          solver_numeric_outputs_post_step,
      )

    # Return redistributed state attributes if triggered, otherwise return
    # unchanged state attributes.
    return jax.lax.cond(
        trigger_sawtooth,
        _redistribute_state,
        lambda: (
            tuple([getattr(core_profiles_t, name) for name in evolving_names]),
            core_sources_t,
            base_conductivity.Conductivity(
                sigma=core_profiles_t.sigma,
                sigma_face=core_profiles_t.sigma_face),
            core_transport_t,
            state.SolverNumericOutputs(),
        ),
    )

  def __hash__(self) -> int:
    return hash((
        self._trigger_model,
        self._redistribution_model,
        self.static_runtime_params_slice,
        self.transport_model,
        self.pedestal_model,
        self.source_models,
    ))

  def __eq__(self, other: object) -> bool:
    return (
        isinstance(other, SawtoothModel)
        and self._trigger_model == other._trigger_model
        and self._redistribution_model == other._redistribution_model
        and self.static_runtime_params_slice
        == other.static_runtime_params_slice
        and self.transport_model == other.transport_model
        and self.pedestal_model == other.pedestal_model
        and self.source_models == other.source_models
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The `enums` module.

Enums shared through the `fvm` package.
"""
import enum


@enum.unique
class InitialGuessMode(enum.Enum):
  """Modes for initial guess of x_new for iterative solvers."""

  # Initialize x_new = x_old
  X_OLD = 0

  # Use the linear solver to guess x_new
  LINEAR = 1



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gets a set of default sources and configurations."""
from typing import Any


def get_default_source_config() -> dict[str, Any]:
  """Returns default sources and configurations.

  This set of sources and params are used by most of the TORAX test
  configurations, including ITER-inspired configs, with additional changes to
  their runtime configurations on top.

  To use, load the dict, update as needed, and then convert to a Sources object:
  .. code-block:: python

    sources_dict = get_default_source_config()
    sources_dict = sources_dict['source_model_config']
    sources_dict['ei_exchange']['Qei_multiplier'] = 0.0
    sources_dict['generic_heat']['P_total'] = 0.0
    sources_dict['fusion']['mode'] = source_runtime_params.Mode.ZERO
    sources_dict['ohmic']['mode'] = source_runtime_params.Mode.ZERO
    default_sources = sources_pydantic_model.Sources.from_dict(sources_dict)
  """
  names = [
      # Current sources (for psi equation)
      'generic_current',
      # Electron density sources/sink (for the n_e equation).
      'generic_particle',
      'gas_puff',
      'pellet',
      # Ion and electron heat sources (for the temp-ion and temp-el eqs).
      'generic_heat',
      'fusion',
      'ei_exchange',
      # Ohmic heat source
      'ohmic',
      'bremsstrahlung',
  ]
  return {name: {} for name in names}



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic which controls the stepping over time of the simulation."""

import dataclasses
import functools
import jax
import jax.numpy as jnp
from torax._src import jax_utils
from torax._src import state
from torax._src.config import build_runtime_params
from torax._src.config import runtime_params_slice
from torax._src.core_profiles import updaters
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.geometry import geometry_provider as geometry_provider_lib
from torax._src.mhd import base as mhd_base
from torax._src.mhd.sawtooth import sawtooth_model as sawtooth_model_lib
from torax._src.orchestration import sim_state
from torax._src.output_tools import post_processing
from torax._src.pedestal_model import pedestal_model as pedestal_model_lib
from torax._src.solver import solver as solver_lib
from torax._src.sources import source_profile_builders
from torax._src.sources import source_profiles as source_profiles_lib
from torax._src.time_step_calculator import time_step_calculator as ts
from torax._src.transport_model import transport_model as transport_model_lib


def _check_for_errors(
    output_state: sim_state.ToraxSimState,
    post_processed_outputs: post_processing.PostProcessedOutputs,
) -> state.SimError:
  """Checks for errors in the simulation state."""
  state_error = output_state.check_for_errors()
  if state_error != state.SimError.NO_ERROR:
    return state_error
  else:
    return post_processed_outputs.check_for_errors()


class SimulationStepFn:
  """Advances the TORAX simulation one time step.

  Unlike the Solver class, which updates certain parts of the state, a
  SimulationStepFn takes in the ToraxSimState and outputs the updated
  ToraxSimState, which contains not only the CoreProfiles but also extra
  simulation state useful for stepping as well as extra outputs useful for
  inspection inside the main run loop in `run_simulation()`. It wraps calls to
  Solver with useful features to increase robustness for convergence, like
  dt-backtracking.
  """

  def __init__(
      self,
      solver: solver_lib.Solver,
      time_step_calculator: ts.TimeStepCalculator,
      transport_model: transport_model_lib.TransportModel,
      pedestal_model: pedestal_model_lib.PedestalModel,
      mhd_models: mhd_base.MHDModels,
  ):
    """Initializes the SimulationStepFn.

    Args:
      solver: Evolves the core profiles.
      time_step_calculator: Calculates the dt for each time step.
      transport_model: Calculates diffusion and convection coefficients.
      pedestal_model: Calculates pedestal coefficients.
      mhd_models: Collection of MHD models applied, e.g. sawtooth
    """
    self._solver = solver
    self._time_step_calculator = time_step_calculator
    self._transport_model = transport_model
    self._pedestal_model = pedestal_model
    self._mhd_models = mhd_models

  @property
  def pedestal_model(self) -> pedestal_model_lib.PedestalModel:
    return self._pedestal_model

  @property
  def solver(self) -> solver_lib.Solver:
    return self._solver

  @property
  def transport_model(self) -> transport_model_lib.TransportModel:
    return self._transport_model

  @property
  def mhd_models(self) -> mhd_base.MHDModels:
    return self._mhd_models

  @property
  def time_step_calculator(self) -> ts.TimeStepCalculator:
    return self._time_step_calculator

  def __call__(
      self,
      static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
      dynamic_runtime_params_slice_provider: build_runtime_params.DynamicRuntimeParamsSliceProvider,
      geometry_provider: geometry_provider_lib.GeometryProvider,
      input_state: sim_state.ToraxSimState,
      previous_post_processed_outputs: post_processing.PostProcessedOutputs,
  ) -> tuple[
      sim_state.ToraxSimState,
      post_processing.PostProcessedOutputs,
      state.SimError,
  ]:
    """Advances the simulation state one time step.

      If a sawtooth model is provided, it will be checked to see if a sawtooth
    should trigger. If it does, the sawtooth model will be applied and instead
    of a full PDE solve, the step_fn will return early with a state following
    sawtooth redistribution, at a t+dt set by the sawtooth model.

    Args:
      static_runtime_params_slice: Static parameters that, if they change,
        should trigger a recompilation of the SimulationStepFn.
      dynamic_runtime_params_slice_provider: Object that returns a set of
        runtime parameters which may change from time step to time step or
        simulation run to run. If these runtime parameters change, it does NOT
        trigger a JAX recompilation.
      geometry_provider: Provides the magnetic geometry for each time step based
        on the ToraxSimState at the start of the time step. The geometry may
        change from time step to time step, so the sim needs a function to
        provide which geometry to use for a given time step. A GeometryProvider
        is any callable (class or function) which takes the ToraxSimState at the
        start of a time step and returns the Geometry for that time step. For
        most use cases, only the time will be relevant from the ToraxSimState
        (in order to support time-dependent geometries).
      input_state: State at the start of the time step, including the core
        profiles which are being evolved.
      previous_post_processed_outputs: Post-processed outputs from the previous
        time step.

    Returns:
      ToraxSimState containing:
        - the core profiles at the end of the time step.
        - time and time step calculator state info.
        - core_sources and core_transport at the end of the time step.
        - solver_numeric_outputs. This contains the number of iterations
          performed in the solver and the error state. The error states are:
            0 if solver converged with fine tolerance for this step
            1 if solver did not converge for this step (was above coarse tol)
            2 if solver converged within coarse tolerance. Allowed to pass with
              a warning. Occasional error=2 has low impact on final sim state.
      PostProcessedOutputs containing:
        - post-processed outputs at the end of the time step.
        - cumulative quantities.
      SimError indicating if an error has occurred during simulation.
    """
    dynamic_runtime_params_slice_t, geo_t = (
        build_runtime_params.get_consistent_dynamic_runtime_params_slice_and_geometry(
            t=input_state.t,
            dynamic_runtime_params_slice_provider=dynamic_runtime_params_slice_provider,
            geometry_provider=geometry_provider,
        )
    )

    # This only computes sources set to explicit in the
    # DynamicSourceConfigSlice. All implicit sources will have their profiles
    # set to 0.
    explicit_source_profiles = source_profile_builders.build_source_profiles(
        dynamic_runtime_params_slice=dynamic_runtime_params_slice_t,
        static_runtime_params_slice=static_runtime_params_slice,
        geo=geo_t,
        core_profiles=input_state.core_profiles,
        source_models=self.solver.source_models,
        explicit=True,
    )

    # If a sawtooth model is provided, and there was no previous
    # sawtooth crash, it will be checked to see if a sawtooth
    # should trigger. If it does, the sawtooth model will be applied and instead
    # of a full PDE solve, the step_fn will return early with a state following
    # sawtooth redistribution, at a t+dt set by the sawtooth model
    # configuration.
    if (self.mhd_models.sawtooth is not None) and (
        not input_state.solver_numeric_outputs.sawtooth_crash
    ):
      assert dynamic_runtime_params_slice_t.mhd.sawtooth is not None
      dt_crash = dynamic_runtime_params_slice_t.mhd.sawtooth.crash_step_duration
      dynamic_runtime_params_slice_t_plus_crash_dt, geo_t_plus_crash_dt = (
          build_runtime_params.get_consistent_dynamic_runtime_params_slice_and_geometry(
              t=input_state.t + dt_crash,
              dynamic_runtime_params_slice_provider=dynamic_runtime_params_slice_provider,
              geometry_provider=geometry_provider,
          )
      )
      # If no sawtooth crash is triggered, output_state and
      # post_processed_outputs will be the same as the input state and
      # previous_post_processed_outputs.
      output_state, post_processed_outputs = _sawtooth_step(
          sawtooth_solver=self.mhd_models.sawtooth,
          static_runtime_params_slice=static_runtime_params_slice,
          dynamic_runtime_params_slice_t=dynamic_runtime_params_slice_t,
          dynamic_runtime_params_slice_t_plus_crash_dt=dynamic_runtime_params_slice_t_plus_crash_dt,
          geo_t=geo_t,
          geo_t_plus_crash_dt=geo_t_plus_crash_dt,
          explicit_source_profiles=explicit_source_profiles,
          input_state=input_state,
          input_post_processed_outputs=previous_post_processed_outputs,
      )
      # If a sawtooth crash was carried out, we exit early with the post-crash
      # state, post-processed outputs, and the error state.
      if output_state.solver_numeric_outputs.sawtooth_crash:
        error_state = _check_for_errors(output_state, post_processed_outputs)
        return output_state, post_processed_outputs, error_state

    dt = self.init_time_step_calculator(
        dynamic_runtime_params_slice_t,
        geo_t,
        input_state,
    )

    # The solver needs the geo and dynamic_runtime_params_slice at time t + dt
    # for implicit computations in the solver. Once geo_t_plus_dt is calculated
    # we can use it to calculate Phibdot for both geo_t and geo_t_plus_dt, which
    # then update the initialized Phibdot=0 in the geo instances.
    dynamic_runtime_params_slice_t_plus_dt, geo_t, geo_t_plus_dt = (
        _get_geo_and_dynamic_runtime_params_at_t_plus_dt_and_phibdot(
            input_state.t,
            dt,
            dynamic_runtime_params_slice_provider,
            geo_t,
            geometry_provider,
        )
    )

    x_new, intermediate_state = self.step(
        dt,
        static_runtime_params_slice,
        dynamic_runtime_params_slice_t,
        dynamic_runtime_params_slice_t_plus_dt,
        geo_t,
        geo_t_plus_dt,
        input_state,
        explicit_source_profiles,
    )

    if static_runtime_params_slice.adaptive_dt:
      # This is a no-op if
      # output_state.solver_numeric_outputs.solver_error_state == 0.
      x_new, intermediate_state, dynamic_runtime_params_slice_t_plus_dt = (
          self.adaptive_step(
              x_new,
              intermediate_state,
              static_runtime_params_slice,
              dynamic_runtime_params_slice_t,
              dynamic_runtime_params_slice_t_plus_dt,
              dynamic_runtime_params_slice_provider,
              geo_t,
              geometry_provider,
              input_state,
              explicit_source_profiles,
          )
      )

    output_state, post_processed_outputs = _finalize_outputs(
        x_new=x_new,
        static_runtime_params_slice=self.solver.static_runtime_params_slice,
        dynamic_runtime_params_slice_t_plus_dt=dynamic_runtime_params_slice_t_plus_dt,
        core_profiles_t=input_state.core_profiles,
        intermediate_state=intermediate_state,
        evolving_names=self.solver.evolving_names,
        input_post_processed_outputs=previous_post_processed_outputs,
    )

    return (
        output_state,
        post_processed_outputs,
        _check_for_errors(output_state, post_processed_outputs),
    )

  def init_time_step_calculator(
      self,
      dynamic_runtime_params_slice_t: runtime_params_slice.DynamicRuntimeParamsSlice,
      geo_t: geometry.Geometry,
      input_state: sim_state.ToraxSimState,
  ) -> jnp.ndarray:
    """First phase: Initialize the solver state.

    Args:
      dynamic_runtime_params_slice_t: Runtime parameters at time t.
      geo_t: The geometry of the torus during this time step of the simulation.
        While the geometry may change, any changes to the grid size can trigger
        recompilation of the solver (if it is jitted) or an error (assuming it
        is JAX-compiled and lowered).
      input_state: State at the start of the time step, including the core
        profiles which are being evolved.

    Returns:
      Time step duration (dt)
    """
    # TODO(b/335598388): We call the transport model both here and in the
    # Solver / CoeffsCallback. We should still refactor the design to more
    # explicitly calculate transport coeffs at delta_t = 0 in only one place,
    # so that we have some flexibility in where to place the jit boundaries.
    transport_coeffs = _calculate_transport_coeffs(
        self.pedestal_model,
        self.transport_model,
        dynamic_runtime_params_slice_t,
        geo_t,
        input_state.core_profiles,
    )

    # initialize new dt and reset solver iterations.
    dt = self._time_step_calculator.next_dt(
        dynamic_runtime_params_slice_t,
        geo_t,
        input_state.core_profiles,
        transport_coeffs,
    )

    crosses_t_final = (
        input_state.t < dynamic_runtime_params_slice_t.numerics.t_final
    ) * (
        input_state.t + input_state.dt
        > dynamic_runtime_params_slice_t.numerics.t_final
    )
    dt = jnp.where(
        jnp.logical_and(
            dynamic_runtime_params_slice_t.numerics.exact_t_final,
            crosses_t_final,
        ),
        dynamic_runtime_params_slice_t.numerics.t_final - input_state.t,
        dt,
    )
    if jnp.any(jnp.isnan(dt)):
      raise ValueError('dt is NaN.')

    return dt

  def step(
      self,
      dt: jnp.ndarray,
      static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
      dynamic_runtime_params_slice_t: runtime_params_slice.DynamicRuntimeParamsSlice,
      dynamic_runtime_params_slice_t_plus_dt: runtime_params_slice.DynamicRuntimeParamsSlice,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      input_state: sim_state.ToraxSimState,
      explicit_source_profiles: source_profiles_lib.SourceProfiles,
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      sim_state.ToraxSimState,
  ]:
    """Performs a simulation step with given dt.

    Solver may fail to converge in which case adaptive_step() can be used to
    try smaller time step durations.

    Args:
      dt: Time step duration.
      static_runtime_params_slice: Static parameters that, if they change,
        should trigger a recompilation of the SimulationStepFn.
      dynamic_runtime_params_slice_t: Runtime parameters at time t.
      dynamic_runtime_params_slice_t_plus_dt: Runtime parameters at time t + dt.
      geo_t: The geometry of the torus during this time step of the simulation.
      geo_t_plus_dt: The geometry of the torus during the next time step of the
        simulation.
      input_state: State at the start of the time step, including the core
        profiles which are being evolved.
      explicit_source_profiles: Explicit source profiles computed based on the
        core profiles at the start of the time step.

    Returns:
      tuple:
        tuple of CellVariables corresponding to the evolved state variables
        An intermediate ToraxSimState with all attributes updated apart from
          core_profiles.
    """

    core_profiles_t = input_state.core_profiles

    # Construct the CoreProfiles object for time t+dt with evolving boundary
    # conditions and time-dependent prescribed profiles not directly solved by
    # PDE system.
    core_profiles_t_plus_dt = updaters.provide_core_profiles_t_plus_dt(
        dt=dt,
        static_runtime_params_slice=static_runtime_params_slice,
        dynamic_runtime_params_slice_t=dynamic_runtime_params_slice_t,
        dynamic_runtime_params_slice_t_plus_dt=dynamic_runtime_params_slice_t_plus_dt,
        geo_t_plus_dt=geo_t_plus_dt,
        core_profiles_t=core_profiles_t,
    )

    # Initial trial for solver. If did not converge (can happen for nonlinear
    # step with large dt) we apply the adaptive time step routine if requested.
    x_new, intermediate_state = self._solver(
        t=input_state.t,
        dt=dt,
        static_runtime_params_slice=static_runtime_params_slice,
        dynamic_runtime_params_slice_t=dynamic_runtime_params_slice_t,
        dynamic_runtime_params_slice_t_plus_dt=dynamic_runtime_params_slice_t_plus_dt,
        geo_t=geo_t,
        geo_t_plus_dt=geo_t_plus_dt,
        core_profiles_t=core_profiles_t,
        core_profiles_t_plus_dt=core_profiles_t_plus_dt,
        core_sources_t=input_state.core_sources,
        core_transport_t=input_state.core_transport,
        explicit_source_profiles=explicit_source_profiles,
    )
    intermediate_state = dataclasses.replace(
        intermediate_state,
        solver_numeric_outputs=dataclasses.replace(
            intermediate_state.solver_numeric_outputs,
            outer_solver_iterations=1,
        ),
    )

    return x_new, intermediate_state

  def adaptive_step(
      self,
      x_old: tuple[cell_variable.CellVariable, ...],
      intermediate_state: sim_state.ToraxSimState,
      static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
      dynamic_runtime_params_slice_t: runtime_params_slice.DynamicRuntimeParamsSlice,
      dynamic_runtime_params_slice_t_plus_dt: runtime_params_slice.DynamicRuntimeParamsSlice,
      dynamic_runtime_params_slice_provider: build_runtime_params.DynamicRuntimeParamsSliceProvider,
      geo_t: geometry.Geometry,
      geometry_provider: geometry_provider_lib.GeometryProvider,
      input_state: sim_state.ToraxSimState,
      explicit_source_profiles: source_profiles_lib.SourceProfiles,
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      sim_state.ToraxSimState,
      runtime_params_slice.DynamicRuntimeParamsSlice,
  ]:
    """Performs adaptive time stepping until solver converges.

    If the initial step has converged (i.e.
    output_state.solver_numeric_outputs.solver_error_state == 0), this
    function is a no-op.

    Args:
      x_old: Tuple containing previous guess of the cell-grid values of the
        evolving variables
      intermediate_state: Partially-solved state after a full step.
      static_runtime_params_slice: Static parameters that, if they change,
        should trigger a recompilation of the SimulationStepFn.
      dynamic_runtime_params_slice_t: Runtime parameters at time t.
      dynamic_runtime_params_slice_t_plus_dt: Runtime parameters at time t + dt.
        Used if a no-op and the original dynamic runtime params slice is
        returned.
      dynamic_runtime_params_slice_provider: Runtime parameters slice provider.
      geo_t: The geometry of the torus during this time step of the simulation.
      geometry_provider: Provides geometry during the next time step of the
        simulation.
      input_state: State at the start of the time step, including the core
        profiles which are being evolved.
      explicit_source_profiles: Explicit source profiles computed based on the
        core profiles at the start of the time step.

    Returns:
      A tuple containing:
        - tuple of CellVariables corresponding to the evolved state variables
        - An intermediate ToraxSimState after adaptive time stepping, complete
          apart from the final core_profiles
        - Dynamic runtime params at time t + dt, where dt is the actual time
          step used.
    """
    core_profiles_t = input_state.core_profiles

    # Check if solver converged. If not, proceed to body_fun
    def cond_fun(
        updated_output: tuple[
            tuple[cell_variable.CellVariable, ...],
            sim_state.ToraxSimState,
            runtime_params_slice.DynamicRuntimeParamsSlice,
        ],
    ) -> bool:
      if updated_output[1].solver_numeric_outputs.solver_error_state == 1:
        do_dt_backtrack = True
      else:
        do_dt_backtrack = False
      return do_dt_backtrack

    # Make a new step with a smaller dt, starting with the original core
    # profiles.
    # Exit if dt < min_dt
    def body_fun(
        updated_output: tuple[
            tuple[cell_variable.CellVariable, ...],
            sim_state.ToraxSimState,
            runtime_params_slice.DynamicRuntimeParamsSlice,
        ],
    ) -> tuple[
        tuple[cell_variable.CellVariable, ...],
        sim_state.ToraxSimState,
        runtime_params_slice.DynamicRuntimeParamsSlice,
    ]:
      _, old_state, old_slice = updated_output
      numerics = old_slice.numerics

      dt = old_state.dt / numerics.dt_reduction_factor
      if jnp.any(jnp.isnan(dt)):
        raise ValueError('dt is NaN.')
      if dt < numerics.min_dt:
        raise ValueError('dt below minimum timestep following adaptation')

      # Calculate dynamic_runtime_params and geo at t + dt.
      # Update geos with phibdot.
      # The updated geo_t is renamed to geo_t_with_phibdot due to name shadowing
      (
          dynamic_runtime_params_slice_t_plus_dt,
          geo_t_with_phibdot,
          geo_t_plus_dt,
      ) = _get_geo_and_dynamic_runtime_params_at_t_plus_dt_and_phibdot(
          input_state.t,
          dt,
          dynamic_runtime_params_slice_provider,
          geo_t,
          geometry_provider,
      )

      core_profiles_t_plus_dt = updaters.provide_core_profiles_t_plus_dt(
          dt=dt,
          static_runtime_params_slice=static_runtime_params_slice,
          dynamic_runtime_params_slice_t=dynamic_runtime_params_slice_t,
          dynamic_runtime_params_slice_t_plus_dt=dynamic_runtime_params_slice_t_plus_dt,
          geo_t_plus_dt=geo_t_plus_dt,
          core_profiles_t=core_profiles_t,
      )
      # The solver returned state is still "intermediate" since the CoreProfiles
      # need to be updated by the evolved CellVariables in x_new
      x_new, intermediate_state = self._solver(
          t=input_state.t,
          dt=dt,
          static_runtime_params_slice=static_runtime_params_slice,
          dynamic_runtime_params_slice_t=dynamic_runtime_params_slice_t,
          dynamic_runtime_params_slice_t_plus_dt=dynamic_runtime_params_slice_t_plus_dt,
          geo_t=geo_t_with_phibdot,
          geo_t_plus_dt=geo_t_plus_dt,
          core_profiles_t=core_profiles_t,
          core_profiles_t_plus_dt=core_profiles_t_plus_dt,
          core_sources_t=input_state.core_sources,
          core_transport_t=input_state.core_transport,
          explicit_source_profiles=explicit_source_profiles,
      )
      intermediate_state = dataclasses.replace(
          intermediate_state,
          solver_numeric_outputs=dataclasses.replace(
              intermediate_state.solver_numeric_outputs,
              outer_solver_iterations=old_state.solver_numeric_outputs.outer_solver_iterations
              + 1,
              inner_solver_iterations=old_state.solver_numeric_outputs.inner_solver_iterations
              + intermediate_state.solver_numeric_outputs.inner_solver_iterations,
          ),
      )

      return x_new, intermediate_state, dynamic_runtime_params_slice_t_plus_dt

    # Iteratively apply the adaptive time step until the solver converges.
    # If the solver has already converged, then the body_fun will not be
    # called and the output will be returned unchanged.
    x_new, intermediate_state, dynamic_runtime_params_slice_t_plus_dt = (
        jax_utils.py_while(
            cond_fun,
            body_fun,
            (x_old, intermediate_state, dynamic_runtime_params_slice_t_plus_dt),
        )
    )

    return x_new, intermediate_state, dynamic_runtime_params_slice_t_plus_dt


@functools.partial(
    jax_utils.jit,
    static_argnames=[
        'sawtooth_solver',
        'static_runtime_params_slice',
    ],
)
def _sawtooth_step(
    *,
    sawtooth_solver: sawtooth_model_lib.SawtoothModel | None,
    static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
    dynamic_runtime_params_slice_t: runtime_params_slice.DynamicRuntimeParamsSlice,
    dynamic_runtime_params_slice_t_plus_crash_dt: runtime_params_slice.DynamicRuntimeParamsSlice,
    geo_t: geometry.Geometry,
    geo_t_plus_crash_dt: geometry.Geometry,
    explicit_source_profiles: source_profiles_lib.SourceProfiles,
    input_state: sim_state.ToraxSimState,
    input_post_processed_outputs: post_processing.PostProcessedOutputs,
) -> tuple[sim_state.ToraxSimState, post_processing.PostProcessedOutputs]:
  """Checks for and handles a sawtooth crash.

  If a sawtooth model is provided and a crash is triggered, this method
  computes the post-crash state and returns it. Otherwise, returns the input
  state and post-processed outputs unchanged.

  Consecutive sawtooth crashes are not allowed since standard PDE steps
  may then not take place. Therefore if the input state has sawtooth_crash set
  to True, then no crash is triggered.

  Args:
    sawtooth_solver: Sawtooth model which carries out sawtooth step..
    static_runtime_params_slice: Static parameters.
    dynamic_runtime_params_slice_t: Dynamic slice at time t.
    dynamic_runtime_params_slice_t_plus_crash_dt: Dynamic slice at time t +
      crash_dt.
    geo_t: Geometry at time t.
    geo_t_plus_crash_dt: Geometry at time t + crash_dt.
    explicit_source_profiles: Explicit source profiles at time t.
    input_state: State at the start of the time step.
    input_post_processed_outputs: Post-processed outputs from the previous step.

  Returns:
    Returns a tuple (output_state, post_processed_outputs).
  """

  # Asserts needed for linter.
  assert dynamic_runtime_params_slice_t.mhd.sawtooth is not None
  assert sawtooth_solver is not None
  dt_crash = dynamic_runtime_params_slice_t.mhd.sawtooth.crash_step_duration

  # Prepare core_profiles_t_plus_crash_dt with new boundary conditions
  # and prescribed profiles if present.
  core_profiles_t_plus_crash_dt = updaters.provide_core_profiles_t_plus_dt(
      dt=dt_crash,
      static_runtime_params_slice=static_runtime_params_slice,
      dynamic_runtime_params_slice_t=dynamic_runtime_params_slice_t,
      dynamic_runtime_params_slice_t_plus_dt=dynamic_runtime_params_slice_t_plus_crash_dt,
      geo_t_plus_dt=geo_t_plus_crash_dt,
      core_profiles_t=input_state.core_profiles,
  )

  (
      x_candidate,
      intermediate_state_candidate,
  ) = sawtooth_solver(
      t=input_state.t,
      dt=dt_crash,
      static_runtime_params_slice=static_runtime_params_slice,
      dynamic_runtime_params_slice_t=dynamic_runtime_params_slice_t,
      dynamic_runtime_params_slice_t_plus_dt=dynamic_runtime_params_slice_t_plus_crash_dt,
      geo_t=geo_t,
      geo_t_plus_dt=geo_t_plus_crash_dt,
      core_profiles_t=input_state.core_profiles,
      core_profiles_t_plus_dt=core_profiles_t_plus_crash_dt,
      core_sources_t=input_state.core_sources,
      core_transport_t=input_state.core_transport,
      explicit_source_profiles=explicit_source_profiles,
  )

  def _make_post_crash_state_and_post_processed_outputs():
    """Returns the post-crash state and post-processed outputs."""
    return _finalize_outputs(
        x_new=x_candidate,
        static_runtime_params_slice=static_runtime_params_slice,
        dynamic_runtime_params_slice_t_plus_dt=dynamic_runtime_params_slice_t_plus_crash_dt,
        core_profiles_t=input_state.core_profiles,
        intermediate_state=intermediate_state_candidate,
        evolving_names=sawtooth_solver.evolving_names,
        input_post_processed_outputs=input_post_processed_outputs,
    )

  return jax.lax.cond(
      intermediate_state_candidate.solver_numeric_outputs.sawtooth_crash,
      _make_post_crash_state_and_post_processed_outputs,
      lambda: (
          input_state,
          input_post_processed_outputs,
      ),
  )


@functools.partial(jax_utils.jit, static_argnums=(0, 1))
def _calculate_transport_coeffs(
    pedestal_model: pedestal_model_lib.PedestalModel,
    transport_model: transport_model_lib.TransportModel,
    dynamic_runtime_params_slice_t: runtime_params_slice.DynamicRuntimeParamsSlice,
    geo_t: geometry.Geometry,
    core_profiles_t: state.CoreProfiles,
) -> state.CoreTransport:
  """Calculates the transport coefficients."""
  pedestal_model_output = pedestal_model(
      dynamic_runtime_params_slice_t, geo_t, core_profiles_t
  )
  return transport_model(
      dynamic_runtime_params_slice_t,
      geo_t,
      core_profiles_t,
      pedestal_model_output,
  )


def _get_geo_and_dynamic_runtime_params_at_t_plus_dt_and_phibdot(
    t: jnp.ndarray,
    dt: jnp.ndarray,
    dynamic_runtime_params_slice_provider: build_runtime_params.DynamicRuntimeParamsSliceProvider,
    geo_t: geometry.Geometry,
    geometry_provider: geometry_provider_lib.GeometryProvider,
) -> tuple[
    runtime_params_slice.DynamicRuntimeParamsSlice,
    geometry.Geometry,
    geometry.Geometry,
]:
  """Returns the geos including Phibdot, and dynamic runtime params at t + dt.

  Args:
    t: Time at which the simulation is currently at.
    dt: Time step duration.
    dynamic_runtime_params_slice_provider: Object that returns a set of runtime
      parameters which may change from time step to time step or simulation run
      to run. If these runtime parameters change, it does NOT trigger a JAX
      recompilation.
    geo_t: The geometry of the torus during this time step of the simulation.
    geometry_provider: Provides the magnetic geometry for each time step based
      on the ToraxSimState at the start of the time step.

  Returns:
    Tuple containing:
      - The dynamic runtime params at time t + dt.
      - The geometry of the torus during this time step of the simulation.
      - The geometry of the torus during the next time step of the simulation.
  """
  dynamic_runtime_params_slice_t_plus_dt, geo_t_plus_dt = (
      build_runtime_params.get_consistent_dynamic_runtime_params_slice_and_geometry(
          t=t + dt,
          dynamic_runtime_params_slice_provider=dynamic_runtime_params_slice_provider,
          geometry_provider=geometry_provider,
      )
  )
  if dynamic_runtime_params_slice_t_plus_dt.numerics.calcphibdot:
    geo_t, geo_t_plus_dt = geometry.update_geometries_with_Phibdot(
        dt=dt,
        geo_t=geo_t,
        geo_t_plus_dt=geo_t_plus_dt,
    )

  return (
      dynamic_runtime_params_slice_t_plus_dt,
      geo_t,
      geo_t_plus_dt,
  )


def _finalize_outputs(
    x_new: tuple[cell_variable.CellVariable, ...],
    static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
    dynamic_runtime_params_slice_t_plus_dt: runtime_params_slice.DynamicRuntimeParamsSlice,
    core_profiles_t: state.CoreProfiles,
    intermediate_state: sim_state.ToraxSimState,
    evolving_names: tuple[str, ...],
    input_post_processed_outputs: post_processing.PostProcessedOutputs,
) -> tuple[sim_state.ToraxSimState, post_processing.PostProcessedOutputs]:
  """Returns the final state and post-processed outputs."""
  final_core_profiles = updaters.update_all_core_profiles_after_step(
      x_new,
      static_runtime_params_slice,
      dynamic_runtime_params_slice_t_plus_dt,
      intermediate_state.geometry,
      intermediate_state.core_sources,
      core_profiles_t,
      intermediate_state.core_profiles,
      evolving_names,
      dt=intermediate_state.dt,
  )
  output_state = dataclasses.replace(
      intermediate_state,
      core_profiles=final_core_profiles,
  )
  post_processed_outputs = post_processing.make_post_processed_outputs(
      sim_state=output_state,
      dynamic_runtime_params_slice=dynamic_runtime_params_slice_t_plus_dt,
      previous_post_processed_outputs=input_post_processed_outputs,
  )
  return output_state, post_processed_outputs



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Profile condition parameters used throughout TORAX simulations."""
import dataclasses

import chex
import pydantic
from torax._src import array_typing
from torax._src.torax_pydantic import torax_pydantic
from typing_extensions import Self
# pylint: disable=invalid-name


@chex.dataclass
class DynamicProfileConditions:
  """Prescribed values and boundary conditions for the core profiles."""

  Ip: array_typing.ScalarFloat
  v_loop_lcfs: array_typing.ScalarFloat
  T_i_right_bc: array_typing.ScalarFloat
  T_e_right_bc: array_typing.ScalarFloat
  # Temperature profiles defined on the cell grid.
  T_e: array_typing.ArrayFloat
  T_i: array_typing.ArrayFloat
  # If provided as array, Psi profile defined on the cell grid.
  psi: array_typing.ArrayFloat | None
  # Electron density profile on the cell grid.
  n_e: array_typing.ArrayFloat
  normalize_n_e_to_nbar: bool
  nbar: array_typing.ScalarFloat
  n_e_nbar_is_fGW: bool
  n_e_right_bc: array_typing.ScalarFloat
  n_e_right_bc_is_fGW: bool
  current_profile_nu: float
  initial_j_is_total_current: bool
  initial_psi_from_j: bool


@chex.dataclass(frozen=True)
class StaticRuntimeParams:
  """Static params for profile conditions."""

  use_v_loop_lcfs_boundary_condition: bool
  normalize_n_e_to_nbar: bool
  # Whether to use absolute ne_bound_right or ne[-1] for setting BC.
  # Set by ne_bound_right condition.
  n_e_right_bc_is_absolute: bool


class ProfileConditions(torax_pydantic.BaseModelFrozen):
  """Generic numeric parameters for the simulation.

  The `from_dict(...)` method can accept a dictionary defined by
  https://torax.readthedocs.io/en/latest/configuration.html#profile-conditions.

  Attributes:
    Ip: Total plasma current in A. Note that if Ip_from_parameters=False in
      geometry, then this Ip will be overwritten by values from the geometry
      data. If use_v_loop_lcfs_boundary_condition, only used as an initial
      condition.
    use_v_loop_lcfs_boundary_condition: Boundary condition at LCFS for Vloop ( =
      dspsi_lcfs/dt ). If use_v_loop_lcfs_boundary_condition is True, then the
      specified Vloop at the LCFS is used as the boundary condition for the psi
      equation; otherwise, Ip is used as the boundary condition.
    v_loop_lcfs: Boundary condition at LCFS for Vloop ( = dpsi_lcfs/dt ).
    T_i_right_bc: Temperature boundary conditions at r=a_minor. If this is
      `None` the boundary condition will instead be taken from `T_i` and `T_e`
      at rhon=1.
    T_e_right_bc: Temperature boundary conditions at r=a_minor. If this is
      `None` the boundary condition will instead be taken from `T_i` and `T_e`
      at rhon=1.
    T_i: Prescribed or evolving values for temperature at different times.
    T_e: Prescribed or evolving values for temperature at different times.
    psi: Initial values for psi. If provided, the initial psi will be taken from
      here. Otherwise, the initial psi will be calculated from either the
      geometry or the "current_profile_nu formula" dependant on the
      `initial_psi_from_j` field.
    n_e: Prescribed or evolving values for electron density at different times.
    normalize_n_e_to_nbar: Whether to renormalize the density profile to have
      the desired line averaged density `nbar`.
    nbar: Line averaged density. In units of m^-3 if n_e_nbar_is_fGW = False. In
      Greenwald fraction if n_e_nbar_is_fGW = True. nGW = Ip/(pi*a^2) with a in
      m, nGW in 10^20 m-3, Ip in MA
    n_e_nbar_is_fGW: Toggle units of nbar
    n_e_right_bc: Density boundary condition for r=a_minor. In units of m^-3 if
      n_e_right_bc_is_fGW = False. In Greenwald fraction if `n_e_right_bc_is_fGW
      = True`. If `n_e_right_bc` is `None` then the boundary condition will
      instead be taken from `n_e` at rho_norm=1. In this case,
      `n_e_right_bc_is_absolute` in the StaticRuntimeParams will be set to
      `False` and n_e_right_bc_is_fGW` will be set to `n_e_nbar_is_fGW`. If
      `n_e_right_bc` is not `None` then `n_e_right_bc_is_absolute` will be set
      to `True`.
    n_e_right_bc_is_fGW: Toggle units of n_e_right_bc.
    current_profile_nu: Peaking factor of "Ohmic" current: j_ohmic = j0*(1 -
      r^2/a^2)^current_profile_nu
    initial_j_is_total_current: Toggles if "Ohmic" current is treated as total
      current upon initialization, or if non-inductive current should be
      included in initial j_total calculation.
    initial_psi_from_j: Toggles if the initial psi calculation is based on the
      "current_profile_nu" current formula, or from the psi available in the
      numerical geometry file. This setting is ignored for the ad-hoc circular
      geometry, which has no numerical geometry.
  """

  Ip: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(15e6)
  use_v_loop_lcfs_boundary_condition: bool = False
  v_loop_lcfs: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.0)
  )
  T_i_right_bc: torax_pydantic.PositiveTimeVaryingScalar | None = None
  T_e_right_bc: torax_pydantic.PositiveTimeVaryingScalar | None = None
  T_i: torax_pydantic.PositiveTimeVaryingArray = (
      torax_pydantic.ValidatedDefault({0: {0: 15.0, 1: 1.0}})
  )
  T_e: torax_pydantic.PositiveTimeVaryingArray = (
      torax_pydantic.ValidatedDefault({0: {0: 15.0, 1: 1.0}})
  )
  psi: torax_pydantic.TimeVaryingArray | None = None
  n_e: torax_pydantic.PositiveTimeVaryingArray = (
      torax_pydantic.ValidatedDefault({0: {0: 1.2e20, 1: 0.8e20}})
  )
  normalize_n_e_to_nbar: bool = False
  nbar: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(
      0.85e20
  )
  n_e_nbar_is_fGW: bool = False
  n_e_right_bc: torax_pydantic.TimeVaryingScalar | None = None
  n_e_right_bc_is_fGW: bool = False
  set_pedestal: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(True)
  )
  current_profile_nu: float = 1.0
  initial_j_is_total_current: bool = False
  initial_psi_from_j: bool = False

  @pydantic.model_validator(mode='after')
  def after_validator(self) -> Self:

    def _sanity_check_profile_boundary_conditions(
        values,
        value_name,
    ):
      """Check that the profile is defined at rho=1.0 for various cases."""
      error_message = (
          f'As no right boundary condition was set for {value_name}, the'
          f' profile for {value_name} must include a rho=1.0 boundary'
          ' condition.'
      )
      if not values.right_boundary_conditions_defined:
        raise ValueError(error_message)

    if self.T_i_right_bc is None:
      _sanity_check_profile_boundary_conditions(self.T_i, 'T_i')
    if self.T_e_right_bc is None:
      _sanity_check_profile_boundary_conditions(self.T_e, 'T_e')
    if self.n_e_right_bc is None:
      _sanity_check_profile_boundary_conditions(self.n_e, 'n_e')
    return self

  def build_dynamic_params(
      self,
      t: chex.Numeric,
  ) -> DynamicProfileConditions:
    """Builds a DynamicProfileConditions."""

    dynamic_params = {
        x.name: getattr(self, x.name)
        for x in dataclasses.fields(DynamicProfileConditions)
    }

    if self.T_e_right_bc is None:
      dynamic_params['T_e_right_bc'] = self.T_e.get_value(
          t, grid_type='face_right'
      )

    if self.T_i_right_bc is None:
      dynamic_params['T_i_right_bc'] = self.T_i.get_value(
          t, grid_type='face_right'
      )

    if self.n_e_right_bc is None:
      dynamic_params['n_e_right_bc'] = self.n_e.get_value(
          t, grid_type='face_right'
      )
      dynamic_params['n_e_right_bc_is_fGW'] = self.n_e_nbar_is_fGW

    def _get_value(x):
      if isinstance(
          x, (torax_pydantic.TimeVaryingScalar, torax_pydantic.TimeVaryingArray)
      ):
        return x.get_value(t)
      else:
        return x

    dynamic_params = {k: _get_value(v) for k, v in dynamic_params.items()}
    return DynamicProfileConditions(**dynamic_params)

  def build_static_params(self) -> StaticRuntimeParams:
    """Builds static runtime params from the config."""
    return StaticRuntimeParams(
        use_v_loop_lcfs_boundary_condition=self.use_v_loop_lcfs_boundary_condition,
        normalize_n_e_to_nbar=self.normalize_n_e_to_nbar,
        n_e_right_bc_is_absolute=False if self.n_e_right_bc is None else True,
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GeometryProvider interface and implementations.

NOTE: Time dependent providers currently live in `geometry.py` and match the
protocol defined here.
"""
from collections.abc import Mapping
import dataclasses
from typing import Protocol, Type

import chex
import numpy as np
from torax._src import interpolated_param
from torax._src import jax_utils
from torax._src.geometry import geometry
from torax._src.torax_pydantic import torax_pydantic
import typing_extensions

# Using invalid-name because we are using the same naming convention as the
# external physics implementations
# pylint: disable=invalid-name


class GeometryProvider(Protocol):
  """Returns the geometry to use during one time step of the simulation.

  A GeometryProvider is any callable (class or function) which takes the
  time of a time step and returns the Geometry for that
  time step. See `SimulationStepFn` for how this callable is used.

  This class is a typing.Protocol, meaning it defines an interface, but any
  function asking for a GeometryProvider as an argument can accept any function
  or class that implements this API without specifically extending this class.

  For instance, the following is an equivalent implementation of the
  ConstantGeometryProvider without actually creating a class, and equally valid.

  .. code-block:: python

    geo = circular_geometry.build_circular_geometry(...)
    constant_geo_provider = lambda t: geo

    def func_expecting_geo_provider(gp: GeometryProvider):
      ... # do something with the provider.

    func_expecting_geo_provider(constant_geo_provider)  # this works.

  NOTE: In order to maintain consistency between the DynamicRuntimeParamsSlice
  and the geometry,
  `build_runtime_params.get_consistent_dynamic_runtime_params_slice_and_geometry`
  should be used to get a Geometry and a corresponding
  DynamicRuntimeParamsSlice.
  """

  def __call__(
      self,
      t: chex.Numeric,
  ) -> geometry.Geometry:
    """Returns the geometry to use during one time step of the simulation.

    The geometry may change from time step to time step, so the sim needs a
    callable to provide which geometry to use for a given time step (this is
    that callable).

    Args:
      t: The time at which the geometry is being requested.

    Returns:
      Geometry of the torus to use for the time step.
    """

  @property
  def torax_mesh(self) -> torax_pydantic.Grid1D:
    """Returns the mesh used by Torax, this is consistent across time."""


class ConstantGeometryProvider(GeometryProvider):
  """Returns the same Geometry for all calls."""

  def __init__(self, geo: geometry.Geometry):
    self._geo = geo

  def __call__(self, t: chex.Numeric) -> geometry.Geometry:
    # The API includes time as an arg even though it is unused in order
    # to match the API of a GeometryProvider.
    del t  # Ignored.
    return self._geo

  @property
  def torax_mesh(self) -> torax_pydantic.Grid1D:
    return self._geo.torax_mesh


@chex.dataclass(frozen=True)
class TimeDependentGeometryProvider:
  """A geometry provider which holds values to interpolate based on time."""

  geometry_type: geometry.GeometryType
  torax_mesh: torax_pydantic.Grid1D
  drho_norm: interpolated_param.InterpolatedVarSingleAxis
  Phi: interpolated_param.InterpolatedVarSingleAxis
  Phi_face: interpolated_param.InterpolatedVarSingleAxis
  R_major: interpolated_param.InterpolatedVarSingleAxis
  a_minor: interpolated_param.InterpolatedVarSingleAxis
  B_0: interpolated_param.InterpolatedVarSingleAxis
  volume: interpolated_param.InterpolatedVarSingleAxis
  volume_face: interpolated_param.InterpolatedVarSingleAxis
  area: interpolated_param.InterpolatedVarSingleAxis
  area_face: interpolated_param.InterpolatedVarSingleAxis
  vpr: interpolated_param.InterpolatedVarSingleAxis
  vpr_face: interpolated_param.InterpolatedVarSingleAxis
  spr: interpolated_param.InterpolatedVarSingleAxis
  spr_face: interpolated_param.InterpolatedVarSingleAxis
  delta_face: interpolated_param.InterpolatedVarSingleAxis
  elongation: interpolated_param.InterpolatedVarSingleAxis
  elongation_face: interpolated_param.InterpolatedVarSingleAxis
  g0: interpolated_param.InterpolatedVarSingleAxis
  g0_face: interpolated_param.InterpolatedVarSingleAxis
  g1: interpolated_param.InterpolatedVarSingleAxis
  g1_face: interpolated_param.InterpolatedVarSingleAxis
  g2: interpolated_param.InterpolatedVarSingleAxis
  g2_face: interpolated_param.InterpolatedVarSingleAxis
  g3: interpolated_param.InterpolatedVarSingleAxis
  g3_face: interpolated_param.InterpolatedVarSingleAxis
  g2g3_over_rhon: interpolated_param.InterpolatedVarSingleAxis
  g2g3_over_rhon_face: interpolated_param.InterpolatedVarSingleAxis
  g2g3_over_rhon_hires: interpolated_param.InterpolatedVarSingleAxis
  F: interpolated_param.InterpolatedVarSingleAxis
  F_face: interpolated_param.InterpolatedVarSingleAxis
  F_hires: interpolated_param.InterpolatedVarSingleAxis
  R_in: interpolated_param.InterpolatedVarSingleAxis
  R_in_face: interpolated_param.InterpolatedVarSingleAxis
  R_out: interpolated_param.InterpolatedVarSingleAxis
  R_out_face: interpolated_param.InterpolatedVarSingleAxis
  spr_hires: interpolated_param.InterpolatedVarSingleAxis
  rho_hires_norm: interpolated_param.InterpolatedVarSingleAxis
  rho_hires: interpolated_param.InterpolatedVarSingleAxis
  _z_magnetic_axis: interpolated_param.InterpolatedVarSingleAxis | None

  @classmethod
  def create_provider(
      cls, geometries: Mapping[float, geometry.Geometry]
  ) -> typing_extensions.Self:
    """Creates a GeometryProvider from a mapping of times to geometries."""
    # Create a list of times and geometries.
    times = np.asarray(list(geometries.keys()), dtype=jax_utils.get_np_dtype())
    geos = list(geometries.values())
    initial_geometry = geos[0]
    for geo in geos:
      if geo.geometry_type != initial_geometry.geometry_type:
        raise ValueError('All geometries must have the same geometry type.')
      if geo.torax_mesh != initial_geometry.torax_mesh:
        raise ValueError('All geometries must have the same mesh.')
    # Create a list of interpolated parameters for each geometry attribute.
    kwargs = {
        'geometry_type': initial_geometry.geometry_type,
        'torax_mesh': initial_geometry.torax_mesh,
    }
    if hasattr(initial_geometry, 'Ip_from_parameters'):
      kwargs['Ip_from_parameters'] = initial_geometry.Ip_from_parameters
    for attr in dataclasses.fields(cls):
      if (
          attr.name == 'geometry_type'
          or attr.name == 'torax_mesh'
          or attr.name == 'Ip_from_parameters'
      ):
        continue
      if attr.name == '_z_magnetic_axis':
        if initial_geometry._z_magnetic_axis is None:  # pylint: disable=protected-access
          kwargs[attr.name] = None
          continue
      kwargs[attr.name] = interpolated_param.InterpolatedVarSingleAxis((
          times,
          np.stack(
              [getattr(g, attr.name) for g in geos],
              axis=0,
              dtype=jax_utils.get_np_dtype(),
          ),
      ))
    return cls(**kwargs)

  def _get_geometry_base(
      self, t: chex.Numeric, geometry_class: Type[geometry.Geometry]
  ):
    """Returns a Geometry instance of the given type at the given time."""
    kwargs = {
        'geometry_type': self.geometry_type,
        'torax_mesh': self.torax_mesh,
    }
    if hasattr(self, 'Ip_from_parameters'):
      kwargs['Ip_from_parameters'] = self.Ip_from_parameters
    for attr in dataclasses.fields(geometry_class):
      if (
          attr.name == 'geometry_type'
          or attr.name == 'torax_mesh'
          or attr.name == 'Ip_from_parameters'
      ):
        continue
      # always initialize Phibdot as zero. It will be replaced once both geo_t
      # and geo_t_plus_dt are provided, and set to be the same for geo_t and
      # geo_t_plus_dt for each given time interval.
      if attr.name == 'Phi_b_dot':
        kwargs[attr.name] = 0.0
        continue
      if attr.name == '_z_magnetic_axis':
        if self._z_magnetic_axis is None:
          kwargs[attr.name] = None
          continue
      kwargs[attr.name] = getattr(self, attr.name).get_value(t)
    return geometry_class(**kwargs)  # pytype: disable=wrong-keyword-args

  def __call__(self, t: chex.Numeric) -> geometry.Geometry:
    """Returns a Geometry instance at the given time."""
    return self._get_geometry_base(t, geometry.Geometry)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class and utils for Qualikiz-based models."""

import chex
from jax import numpy as jnp
from torax._src import constants as constants_module
from torax._src import state
from torax._src.geometry import geometry
from torax._src.physics import collisions
from torax._src.physics import psi_calculations
from torax._src.transport_model import quasilinear_transport_model


@chex.dataclass(frozen=True)
class DynamicRuntimeParams(quasilinear_transport_model.DynamicRuntimeParams):
  """Shared parameters for Qualikiz-based models."""

  collisionality_multiplier: float
  avoid_big_negative_s: bool
  smag_alpha_correction: bool
  q_sawtooth_proxy: bool


# pylint: disable=invalid-name
@chex.dataclass(frozen=True)
class QualikizInputs(quasilinear_transport_model.QuasilinearInputs):
  """Inputs to Qualikiz-based models."""

  Z_eff_face: chex.Array
  q: chex.Array
  smag: chex.Array
  x: chex.Array
  Ti_Te: chex.Array
  log_nu_star_face: chex.Array
  normni: chex.Array
  alpha: chex.Array
  epsilon_lcfs: chex.Array

  # Also define the logarithmic gradients using standard QuaLiKiz notation.
  @property
  def Ati(self) -> chex.Array:
    return self.lref_over_lti

  @property
  def Ate(self) -> chex.Array:
    return self.lref_over_lte

  @property
  def Ane(self) -> chex.Array:
    return self.lref_over_lne

  @property
  def Ani0(self) -> chex.Array:
    return self.lref_over_lni0

  @property
  def Ani1(self) -> chex.Array:
    return self.lref_over_lni1


class QualikizBasedTransportModel(
    quasilinear_transport_model.QuasilinearTransportModel
):
  """Base class for Qualikiz-based transport models."""

  def _prepare_qualikiz_inputs(
      self,
      Z_eff_face: chex.Array,
      density_reference: chex.Numeric,
      transport: DynamicRuntimeParams,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> QualikizInputs:
    """Prepare Qualikiz inputs."""
    constants = constants_module.CONSTANTS

    # define radial coordinate as midplane average r
    # (typical assumption for transport models developed in circular geo)
    rmid = (geo.R_out - geo.R_in) * 0.5
    rmid_face = (geo.R_out_face - geo.R_in_face) * 0.5

    # gyrobohm diffusivity
    # (defined here with Lref=a_minor due to QLKNN training set normalization)
    chiGB = quasilinear_transport_model.calculate_chiGB(
        reference_temperature=core_profiles.T_i.face_value(),
        reference_magnetic_field=geo.B_0,
        reference_mass=core_profiles.A_i,
        reference_length=geo.a_minor,
    )

    # transport coefficients from the qlknn-hyper-10D model
    # (K.L. van de Plassche PoP 2020)

    # set up input vectors (all as jax.numpy arrays on face grid)

    # Calculate normalized logarithmic gradients
    normalized_logarithmic_gradients = quasilinear_transport_model.NormalizedLogarithmicGradients.from_profiles(
        core_profiles=core_profiles,
        radial_coordinate=rmid,
        reference_length=geo.R_major,
    )

    q = core_profiles.q_face

    # Due to QuaLikiz geometry assumptions, we need to calculate s with respect
    # to the midplane average, and not use the standard s_face from CoreProfiles
    smag = psi_calculations.calc_s_rmid(
        geo,
        core_profiles.psi,
    )

    # Inverse aspect ratio at LCFS.
    epsilon_lcfs = rmid_face[-1] / geo.R_major
    # Local normalized radius.
    x = rmid_face / rmid_face[-1]
    x = jnp.where(jnp.abs(x) < constants.eps, constants.eps, x)

    # Ion to electron temperature ratio
    Ti_Te = (
        core_profiles.T_i.face_value() / core_profiles.T_e.face_value()
    )

    # logarithm of normalized collisionality
    nu_star = collisions.calc_nu_star(
        geo=geo,
        core_profiles=core_profiles,
        density_reference=density_reference,
        Z_eff_face=Z_eff_face,
        collisionality_multiplier=transport.collisionality_multiplier,
    )
    log_nu_star_face = jnp.log10(nu_star)

    # calculate alpha for magnetic shear correction (see S. van Mulders NF 2021)
    alpha = quasilinear_transport_model.calculate_alpha(
        core_profiles=core_profiles,
        density_reference=density_reference,
        q=q,
        reference_magnetic_field=geo.B_0,
        normalized_logarithmic_gradients=normalized_logarithmic_gradients,
    )

    # to approximate impact of Shafranov shift. From van Mulders Nucl. Fusion
    # 2021.
    smag = jnp.where(
        transport.smag_alpha_correction,
        smag - alpha / 2,
        smag,
    )

    # very basic ad-hoc sawtooth model
    smag = jnp.where(
        jnp.logical_and(
            transport.q_sawtooth_proxy,
            q < 1,
        ),
        0.1,
        smag,
    )

    q = jnp.where(
        jnp.logical_and(
            transport.q_sawtooth_proxy,
            q < 1,
        ),
        1,
        q,
    )

    smag = jnp.where(
        jnp.logical_and(
            transport.avoid_big_negative_s,
            smag - alpha < -0.2,
        ),
        alpha - 0.2,
        smag,
    )
    normni = core_profiles.n_i.face_value() / core_profiles.n_e.face_value()
    return QualikizInputs(
        Z_eff_face=Z_eff_face,
        lref_over_lti=normalized_logarithmic_gradients.lref_over_lti,
        lref_over_lte=normalized_logarithmic_gradients.lref_over_lte,
        lref_over_lne=normalized_logarithmic_gradients.lref_over_lne,
        lref_over_lni0=normalized_logarithmic_gradients.lref_over_lni0,
        lref_over_lni1=normalized_logarithmic_gradients.lref_over_lni1,
        q=q,
        smag=smag,
        x=x,
        Ti_Te=Ti_Te,
        log_nu_star_face=log_nu_star_face,
        normni=normni,
        chiGB=chiGB,
        Rmaj=geo.R_major,
        Rmin=geo.a_minor,
        alpha=alpha,
        epsilon_lcfs=epsilon_lcfs,
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Default configs for testing."""
from typing import Any


def get_default_config_dict() -> dict[str, Any]:
  return {
      'profile_conditions': {},
      'plasma_composition': {},
      'numerics': {},
      'geometry': {'geometry_type': 'circular', 'n_rho': 4},
      'sources': {},
      'solver': {},
      'transport': {},
      'pedestal': {},
  }



# Copyright 2024 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Common types for using jaxtyping in TORAX."""
import chex
import jaxtyping as jt

ScalarFloat = jt.Float[chex.Array | float, ""]
ScalarBool = jt.Bool[chex.Array | bool, ""]
ScalarInt = jt.Int[chex.Array | int, ""]
ArrayFloat = jt.Float[chex.Array, "rhon"]
ArrayBool = jt.Bool[chex.Array, "rhon"]



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataclass representing runtime parameter inputs to the transport models.

This is the dataclass runtime config exposed to the user. The actual model gets
a time-interpolated version of this config via the DynamicRuntimeParams.
"""
import chex
from torax._src import array_typing
from torax._src import jax_utils


# pylint: disable=invalid-name
@chex.dataclass(frozen=True)
class DynamicRuntimeParams:
  """Input params for the transport model which can be used as compiled args."""

  chi_min: float
  chi_max: float
  D_e_min: float
  D_e_max: float
  V_e_min: float
  V_e_max: float
  apply_inner_patch: array_typing.ScalarBool
  D_e_inner: array_typing.ScalarFloat
  V_e_inner: array_typing.ScalarFloat
  chi_i_inner: array_typing.ScalarFloat
  chi_e_inner: array_typing.ScalarFloat
  rho_inner: array_typing.ScalarFloat
  apply_outer_patch: array_typing.ScalarBool
  D_e_outer: array_typing.ScalarFloat
  V_e_outer: array_typing.ScalarFloat
  chi_i_outer: array_typing.ScalarFloat
  chi_e_outer: array_typing.ScalarFloat
  rho_outer: array_typing.ScalarFloat
  smoothing_width: float
  smooth_everywhere: bool

  def __post_init__(self):
    jax_utils.error_if(
        self.rho_outer,
        self.rho_outer <= self.rho_inner,
        'rho_outer must be greater than rho_inner.',
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper for finding safety factor outputs."""
import functools

import chex
import jax
from jax import numpy as jnp
from torax._src import array_typing


@chex.dataclass(frozen=True)
class SafetyFactorFit:
  """Collection of outputs calculated after each simulation step.

  Attributes:
    rho_q_min: rho_norm at the minimum q.
    q_min: Minimum q value.
    rho_q_3_2_first: First outermost rho_norm value that intercepts the
      q=3/2 plane.
    rho_q_2_1_first: First outermost rho_norm value that intercepts the q=2/1
      plane.
    rho_q_3_1_first: First outermost rho_norm value that intercepts the q=3/1
      plane.
    rho_q_3_2_second: Second outermost rho_norm value that intercepts the
      q=3/2 plane.
    rho_q_2_1_second: Second outermost rho_norm value that intercepts the q=2/1
      plane.
    rho_q_3_1_second: Second outermost rho_norm value that intercepts the q=3/1
      plane.
  """
  rho_q_min: array_typing.ScalarFloat
  q_min: array_typing.ScalarFloat
  rho_q_3_2_first: array_typing.ScalarFloat
  rho_q_2_1_first: array_typing.ScalarFloat
  rho_q_3_1_first: array_typing.ScalarFloat
  rho_q_3_2_second: array_typing.ScalarFloat
  rho_q_2_1_second: array_typing.ScalarFloat
  rho_q_3_1_second: array_typing.ScalarFloat


def _sliding_window_of_three(flat_array: jax.Array) -> jax.Array:
  """Sliding window equivalent to numpy.lib.stride_tricks.sliding_window."""
  window_size = 3
  starts = jnp.arange(len(flat_array) - window_size + 1)
  return jax.vmap(
      lambda start: jax.lax.dynamic_slice(flat_array, (start,), (window_size,))
  )(starts)


def _fit_polynomial_to_intervals_of_three(
    rho_norm: jax.Array, q_face: jax.Array
) -> tuple[jax.Array, jax.Array, jax.Array]:
  """Fit polynomial to every set of three points in given arrays."""
  q_face_intervals = _sliding_window_of_three(
      q_face,
  )
  rho_norm_intervals = _sliding_window_of_three(
      rho_norm,
  )

  @jax.vmap
  def batch_polyfit(
      q_face_interval: jax.Array, rho_norm_interval: jax.Array
  ) -> jax.Array:
    """Solve Ax=b to get coefficients for a quadratic polynomial."""
    chex.assert_shape(q_face_interval, (3,))
    chex.assert_shape(rho_norm_interval, (3,))
    rho_norm_squared = rho_norm_interval**2
    A = jnp.array([  # pylint: disable=invalid-name
        [rho_norm_squared[0], rho_norm_interval[0], 1],
        [rho_norm_squared[1], rho_norm_interval[1], 1],
        [rho_norm_squared[2], rho_norm_interval[2], 1],
    ])
    b = jnp.array([q_face_interval[0], q_face_interval[1], q_face_interval[2]])
    coeffs = jnp.linalg.solve(A, b)
    return coeffs

  return (
      batch_polyfit(q_face_intervals, rho_norm_intervals),
      rho_norm_intervals,
      q_face_intervals,
  )


@jax.vmap
def _minimum_location_value_in_interval(
    coeffs: jax.Array, rho_norm_interval: jax.Array, q_interval: jax.Array
) -> tuple[jax.Array, jax.Array]:
  """Returns the minimum value and location of the fit quadratic in the interval."""
  min_interval, max_interval = rho_norm_interval[0], rho_norm_interval[1]
  q_min_interval, q_max_interval = (
      q_interval[0],
      q_interval[1],
  )
  a, b = coeffs[0], coeffs[1]
  extremum_location = -b / (2 * a)
  extremum_in_interval = jnp.greater(
      extremum_location, min_interval
  ) & jnp.less(extremum_location, max_interval)
  extremum_value = jax.lax.cond(
      extremum_in_interval,
      lambda x: jnp.polyval(coeffs, x),
      lambda x: jnp.inf,
      extremum_location,
  )

  interval_minimum_location, interval_minimum_value = jax.lax.cond(
      jnp.less(q_min_interval, q_max_interval),
      lambda: (min_interval, q_min_interval),
      lambda: (max_interval, q_max_interval),
  )
  overall_minimum_location, overall_minimum_value = jax.lax.cond(
      jnp.less(interval_minimum_value, extremum_value),
      lambda: (interval_minimum_location, interval_minimum_value),
      lambda: (extremum_location, extremum_value),
  )
  return overall_minimum_location, overall_minimum_value


def _find_roots_quadratic(coeffs: jax.Array) -> jax.Array:
  """Finds the roots of a quadratic equation."""
  a, b, c = coeffs[0], coeffs[1], coeffs[2]
  determinant = b**2 - 4.0 * a * c
  roots_exist = jnp.greater(determinant, 0)
  plus_root = jax.lax.cond(
      roots_exist,
      lambda: (-b + jnp.sqrt(determinant)) / (2.0 * a),
      lambda: -jnp.inf,
  )
  minus_root = jax.lax.cond(
      roots_exist,
      lambda: (-b - jnp.sqrt(determinant)) / (2.0 * a),
      lambda: -jnp.inf,
  )
  return jnp.array([plus_root, minus_root])


@functools.partial(jax.vmap, in_axes=(0, 0, None))
def _root_in_interval(
    coeffs: jax.Array, interval: jax.Array, q_surface: float
) -> jax.Array:
  """Finds roots of quadratic(coeffs)=q_surface in the interval."""
  intercept_coeffs = coeffs - jnp.array([0.0, 0.0, q_surface])
  min_interval, max_interval = interval[0], interval[1]
  root_values = _find_roots_quadratic(intercept_coeffs)
  in_interval = jnp.greater(root_values, min_interval) & jnp.less(
      root_values, max_interval
  )
  return jnp.where(in_interval, root_values, -jnp.inf)


@jax.jit
def find_min_q_and_q_surface_intercepts(
    rho_norm: jax.Array, q_face: jax.Array
) -> SafetyFactorFit:
  """Finds the minimum q and the q surface intercepts.

  This method divides the input arrays into groups of three points, fits a
  quadratic to each interval containing three points. It then uses the fit
  quadratics to find the minimum q value as well as any intercepts of the q=3/2,
  q=2/1, and q=3/1 planes.

  As the quadratic fits could have overlapping definitions where the groups
  overlap, we choose the quadratic for the domain [rho_norm[i-1], rho_norm[i]]
  defined using the points (i, i-1, i-2).
  (There is one exception which is the first group for which the domain for
  points in [rho_norm[0], rho_norm[2]] are defined using points (0, 1, 2))

  Args:
    rho_norm: Array of rho_norm values on face grid.
    q_face: Array of q values on face grid.

  Returns:
    rho_q_min: rho_norm at the minimum q.
    q_min: minimum q value.
    outermost_rho_q_3_2: Array of two outermost rho_norm values that intercept
    the q=3/2 plane. If fewer than two intercepts are found, entries will be set
    to -inf.
    outermost_rho_q_2_1: Array of 2 outermost rho_norm values that intercept the
    q=2/1 plane. If fewer than two intercepts are found, entries will be set to
    -inf.
    outermost_rho_q_3_1: Array of 2 outermost rho_norm values that intercept the
    q=3/1 plane. If fewer than two intercepts are found, entries will be set to
    -inf.
  """
  if len(q_face) != len(rho_norm):
    raise ValueError(
        f'Input arrays must have the same length. {len(q_face)} !='
        f' {len(rho_norm)}'
    )
  if len(q_face) < 4:
    raise ValueError('Input arrays must have at least four points.')
  # Sort in case input is not sorted.
  sorted_indices = jnp.argsort(rho_norm)
  rho_norm = rho_norm[sorted_indices]
  q_face = q_face[sorted_indices]
  # Fit polynomial to every set of three points in given arrays.
  poly_coeffs, rho_norm_3, q_face_3 = _fit_polynomial_to_intervals_of_three(
      rho_norm, q_face
  )
  # To bridge across the entire span of rho_norm, and avoid overlapping
  # interval domains, we define the first group from points (0, 2), and all
  # subsequent groups from (i+1, i+2), from i=1 until i=len(rho_norm)-1.
  first_rho_norm = jnp.expand_dims(
      jnp.array([rho_norm[0], rho_norm[2]]), axis=0
  )
  first_q_face = jnp.expand_dims(jnp.array([q_face[0], q_face[2]]), axis=0)
  rho_norms = jnp.concat([first_rho_norm, rho_norm_3[1:, 1:]], axis=0)
  q_faces = jnp.concat([first_q_face, q_face_3[1:, 1:]], axis=0)

  # Find the minimum q value and its location for each interval.
  rho_q_min_intervals, q_min_intervals = _minimum_location_value_in_interval(
      poly_coeffs, rho_norms, q_faces
  )
  # Find the minimum q value and its location out of all intervals.
  arg_q_min = jnp.argmin(q_min_intervals)
  rho_q_min = rho_q_min_intervals[arg_q_min]
  q_min = q_min_intervals[arg_q_min]

  # Find the outermost rho_norm values that intercept the q=3/2, q=2/1, and
  # q=3/1 planes. If none are found, fill from the left with -jnp.inf.
  rho_q_3_2 = _root_in_interval(poly_coeffs, rho_norms, 1.5).flatten()
  outermost_rho_q_3_2 = rho_q_3_2[jnp.argsort(rho_q_3_2)[-2:]]
  rho_q_2_1 = _root_in_interval(poly_coeffs, rho_norms, 2.0).flatten()
  outermost_rho_q_2_1 = rho_q_2_1[jnp.argsort(rho_q_2_1)[-2:]]
  rho_q_3_1 = _root_in_interval(poly_coeffs, rho_norms, 3.0).flatten()
  outermost_rho_q_3_1 = rho_q_3_1[jnp.argsort(rho_q_3_1)[-2:]]

  return SafetyFactorFit(
      rho_q_min=rho_q_min,
      q_min=q_min,
      rho_q_3_2_first=outermost_rho_q_3_2[0],
      rho_q_2_1_first=outermost_rho_q_2_1[0],
      rho_q_3_1_first=outermost_rho_q_3_1[0],
      rho_q_3_2_second=outermost_rho_q_3_2[1],
      rho_q_2_1_second=outermost_rho_q_2_1[1],
      rho_q_3_1_second=outermost_rho_q_3_1[1],
  )
